-- Minimap info, basically just P_InitMinimapInfo converted from C to Lua

rawset(_G, "SG_Minimap", nil)
local minimapinit = false

local function freeMinimap()
    rawset(_G, "SG_Minimap", nil)
    minimapinit = false
end

local function getLevelBBox()
    local bbox = { min_x = nil, min_y = nil, max_x = nil, max_y = nil, }

    for vertex in vertexes.iterate do
        if bbox.min_x == nil or vertex.x < bbox.min_x then
            bbox.min_x = vertex.x
        end

        if bbox.max_x == nil or vertex.x > bbox.max_x then
            bbox.max_x = vertex.x
        end

        if bbox.min_y == nil or vertex.y < bbox.min_y then
            bbox.min_y = vertex.y
        end

        if bbox.max_y == nil or vertex.y > bbox.max_y then
            bbox.max_y = vertex.y
        end
    end

    return bbox
end

local function initMinimap(v)
    minimapinit = true

    local minimapinfo = getLevelBBox()

    local picname = string.format("%sR", G_BuildMapName(gamemap))

    if v.patchExists(picname) then
        minimapinfo.minimap_pic = v.cachePatch(picname)
    end

    -- You might be wondering why these are being bitshift here
    -- it's because mapwidth and height would otherwise overflow for maps larger than half the size possible...
    -- map boundaries and sizes will ALWAYS be whole numbers thankfully
    -- later calculations take into consideration that these are actually not in terms of FRACUNIT though
    --
    -- Changed bitshifts to dividing by fracunit in lua, since bitshifts can behave differently for negative numbers
    -- between c and lua
    minimapinfo.max_x = $ / FRACUNIT
    minimapinfo.min_x = $ / FRACUNIT
    minimapinfo.max_y = $ / FRACUNIT
    minimapinfo.min_y = $ / FRACUNIT
    minimapinfo.map_w = minimapinfo.max_x - minimapinfo.min_x
    minimapinfo.map_h = minimapinfo.max_y - minimapinfo.min_y

    minimapinfo.minimap_w, minimapinfo.minimap_h = 100, 100

    local a = FixedDiv(minimapinfo.minimap_w<<FRACBITS, minimapinfo.map_w<<4)
    local b = FixedDiv(minimapinfo.minimap_h<<FRACBITS, minimapinfo.map_h<<4)

    if a < b then
        minimapinfo.minimap_h = FixedMul(a, minimapinfo.map_h)>>(FRACBITS-4)
        minimapinfo.zoom = a
    else
        if a ~= b then
            minimapinfo.minimap_w = FixedMul(b, minimapinfo.map_w)>>(FRACBITS-4)
        end

        minimapinfo.zoom = b
    end

    minimapinfo.zoom = $ >> (FRACBITS-4)
    minimapinfo.zoom = $ - (minimapinfo.zoom/20)

    -- These should always be small enough to be bitshift back right now
    minimapinfo.offs_x = FixedMul((minimapinfo.min_x + minimapinfo.map_w/2) << FRACBITS, minimapinfo.zoom)
    minimapinfo.offs_y = FixedMul((minimapinfo.min_y + minimapinfo.map_h/2) << FRACBITS, minimapinfo.zoom)

    rawset(_G, "SG_Minimap", minimapinfo)
end

addHook("MapLoad", function()
    -- Can't init it here yet
    freeMinimap()
end)

hud.add(function(v)
    if minimapinit then
        return
    end

    initMinimap(v)
end)
