-- aj@worldsbe.st
-- "FIRMSNEAKER" - boosts sneaker items to 50% speedboost, even on hard, while not affecting dash panels
-- let's move this out of server repacks since file limit is a billion now.

-- Edited for Sunflower's Garden by Indev
-- Firmsneaker only activates if you are in red position or lower
-- This allows further oppertunies for people from the back to catch up back to the middle batch.
-- Not active in Tech or DriftNitro
--
-- Notice: Uses pseudorng.lua for VFX (N_RandomRange)

freeslot("SPR_FIRM")

local cv_firmsneaker = CV_RegisterVar {
    name = "firmsneaker",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,
    flags = CV_NETVAR,
}

local cv_positiontimer = CV_RegisterVar {
    name = "firmsneaker_positiontimer",
    defaultvalue = 2*TICRATE,
    possiblevalue = CV_Natural,
    flags = CV_NETVAR,
}

local cv_minposition = CV_RegisterVar {
    name = "firmsneaker_minposition",
    defaultvalue = 6,
    possiblevalue = CV_Natural,
    flags = CV_NETVAR,
}

local cv_firmsneakervfx = CONFIG_RegisterVar {
    name = "firmsneakervfx",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,

	config_menu = "Other",
	displayname = "Firm Sneaker VFX",
	description = "Display special effect when player can use firm sneaker.",
}

local starttime = 7*TICRATE

local fs = cv_firmsneaker.value

local function thinkFirmsneakerVFX(mo)
    if not (mo.target and mo.target.valid) then
        P_RemoveMobj(mo)
        return
    end

    mo.sprite = SPR_FIRM
    mo.frame = 0
    mo.tics = -1

    local target = mo.target

    P_MoveOrigin(mo, target.x + mo.offsetx, target.y + mo.offsety, target.z + mo.offsetz)

    mo.offsetz = $ + mapobjectscale*((mo.eflags & MFE_VERTICALFLIP) and -4 or 4)

    -- Fade out
    if mo.fuse < 10 then
        local frame = mo.frame & (~FF_TRANSMASK)
        frame = $ | ((10 - max(mo.fuse, 1))<<FF_TRANSSHIFT)
        mo.frame = frame
    end
end

local function updateFirmsneakerVFX(pmo)
    if not pmo.FSvfx then pmo.FSvfx = {} end

    local vfx = pmo.FSvfx

    for _, mo in ipairs(vfx) do
        if mo.valid then
            thinkFirmsneakerVFX(mo)
        end
    end

    for i = #vfx, 1, -1 do
        if not vfx[i].valid then
            table.remove(vfx, i)
        end
    end
end

-- Grrr not exposed
local function P_IsLocalPlayer(p)
    for dp in displayplayers.iterate do
        if p == dp then return true end
    end

    return false
end

rawset(_G, "SG_CanHaveFirmsneaker", function(p)
    return K_IsPlayerLosing(p) and p.kartstuff[k_position] >= cv_minposition.value
end)

local function spawnFirmsneakerVFX(pmo)
    if not cv_firmsneakervfx.value then return end
    if not P_IsLocalPlayer(pmo.player) then return end
    if not pmo.FSvfx then pmo.FSvfx = {} end

    local vfx = pmo.FSvfx

    if leveltime % 6 == 0 then
        local offsetx = N_RandomRange(-30, 30)*mapobjectscale
		local offsety = N_RandomRange(-30, 30)*mapobjectscale
		local offsetz = N_RandomRange(0, 10)*mapobjectscale

		local mo = P_SpawnMobj(pmo.x + offsetx, pmo.y + offsety, pmo.z + offsetz, MT_THOK)
		K_MatchGenericExtraFlags(mo, pmo)
		mo.fuse = 3*TICRATE/5
		mo.target = pmo
		mo.offsetx = offsetx
		mo.offsety = offsety
		mo.offsetz = offsetz
		mo.scale = mapobjectscale/2
		mo.destscale = mapobjectscale
		mo.scalespeed = mapobjectscale/24

        table.insert(vfx, mo)
    end
end

addHook("NetVars", function(net)
    fs = net(fs)
end)

addHook("MapLoad", function()
    fs = cv_firmsneaker.value and not (booststack and booststack.running)

    for p in players.iterate do
        p.FSpositiontimer = 0
    end
end)

addHook("ThinkFrame", function()
    if (not fs) or (leveltime < starttime+30*TICRATE) then return end

    for p in players.iterate do
        if p.spectator then continue end

        if p.FSpositiontimer == nil then p.FSpositiontimer = 0 end

        if SG_CanHaveFirmsneaker(p) then
            p.FSpositiontimer = min(p.FSpositiontimer+1, cv_positiontimer.value)
        elseif p.FSpositiontimer > 0 then
            p.FSpositiontimer = max(p.FSpositiontimer-1, 0)
        end

        if p.FSpositiontimer > 0 then
            if p.kartstuff[k_itemtype] == KITEM_SNEAKER or p.kartstuff[k_itemtype] == KITEM_ROCKETSNEAKER or p.kartstuff[k_rocketsneakertimer] then
                spawnFirmsneakerVFX(p.mo)
            end
        end

        updateFirmsneakerVFX(p.mo)

        local cancelfs = p.cmd.buttons & BT_BACKWARD

        if p.FSlastframeitem ~= nil then
            if p.kartstuff[k_sneakertimer] > p.FSlastframetimer then
                if cancelfs then
                    p.FSdohardsneaker = false
                    p.FSfirmbonus = 0
                elseif p.FSlastframeitem == KITEM_SNEAKER and p.kartstuff[k_itemtype] ~= KITEM_SNEAKER then
                    p.FSdohardsneaker = true
                elseif p.kartstuff[k_itemtype] == KITEM_SNEAKER and p.kartstuff[k_itemamount] < p.FSlastframeamount then
                    p.FSdohardsneaker = true
                elseif p.kartstuff[k_rocketsneakertimer] and p.FSlastframebutton == 0 and (p.cmd.buttons & BT_ATTACK) then
                    p.FSdohardsneaker = true
                else
                    p.FSdohardsneaker = false
                end
            end
        end

        if p.FSpositiontimer > 0 then
            if p.FSdohardsneaker and p.kartstuff[k_sneakertimer] then
                p.kartstuff[k_speedboost] = max($, FRACUNIT/2)
                p.FSfirmbonus = p.kartstuff[k_sneakertimer]
            elseif p.FSfirmbonus and p.kartstuff[k_sneakertimer] then
                p.kartstuff[k_speedboost] = max($, FRACUNIT/2)
            end
        end

        p.FSfirmbonus = $ or 0
        p.FSfirmbonus = max(0, $-1)
        p.FSlastframeitem = p.kartstuff[k_itemtype]
        p.FSlastframeamount = p.kartstuff[k_itemamount]
        p.FSlastframetimer = p.kartstuff[k_sneakertimer]

        if p.FSlastframebutton == nil then p.FSlastframebutton = 0 end

        p.FSlastframebutton = (p.cmd.buttons & BT_ATTACK) and ($+1) or 0
        p.FSdohardsneaker = p.FSdohardsneaker or false
    end
end)

local S_BOOSTFLAME = S_BOOSTFLAME
local SKINCOLOR_BLUEBERRY = SKINCOLOR_BLUEBERRY
local SKINCOLOR_FLAME = SKINCOLOR_FLAME

local function getTargetPlayer(mo, rec)
    if not (mo.target and mo.target.valid) then return end

    if mo.target and mo.target.player then return mo.target.player end

    -- Just check target and target's target. No need for endless recursion
    if not rec then
        return getTargetPlayer(mo.target, true)
    end
end

addHook("MobjThinker", function(mo)
    if not fs then return end

    local p = getTargetPlayer(mo)

    if not p or (p.FSpositiontimer == nil) then return end

    if mo.state == S_BOOSTFLAME then
        mo.colorized = true

        if p.FSpositiontimer > 0 and (p.FSdohardsneaker or p.FSfirmbonus) then
            mo.color = SKINCOLOR_BLUEBERRY
        else
            mo.color = SKINCOLOR_FLAME
        end
    end
end, MT_BOOSTFLAME)

--hud.add(function(v, p, c)
--        if p.FSlastframeitem == nil then return end
--        v.drawString(0, 0, "SPEEDBOOST: "..p.kartstuff[k_speedboost])
--        v.drawString(0, 10, "SNEAKER TIMER: "..p.kartstuff[k_sneakertimer])
--        v.drawString(0, 20, "FIRMBONUS: "..p.FSfirmbonus)
--end)
