local cv_enable = CONFIG_RegisterVar{
    name = "saltyhop_var",
    defaultvalue = "Off",
    PossibleValue = CV_OnOff,
	config_menu = "Other",
	displayname = "Saltyhop",
	description = "Make your character jump when drifting.",
}

local cv_sfx = CONFIG_RegisterVar{
    name = "saltyhop_sfx",
    defaultvalue = "On",
    PossibleValue = CV_OnOff,
	config_menu = "Other",
	displayname = "Saltyhop SFX",
	description = "Play sounds when someone jumps.",
}

local cv_hide = CONFIG_RegisterVar{
    name = "saltyhop_hide",
    defaultvalue = "Off",
    PossibleValue = CV_OnOff,
	config_menu = "Other",
	displayname = "Hide Saltyhop",
	description = "Don't show jumps.",
}

-- ffs why do i have to do that
-- G: because lua scripts are loaded before cvars are initialized, if loaded via -file
local cv_hardcode
addHook("MapLoad", function()
    if cv_hardcode == nil then
        cv_hardcode = CV_FindVar("hardcodehop")
		-- G: don't check again
		cv_hardcode = $ or false
    end
end)

//saltyhop command by SMS Alfredo
COM_AddCommand("saltyhop", function(player, arg)
    arg = arg and arg:lower()

	if arg == "0" or arg == "off" or arg == "false" then
		player.saltyhop = false
		if player == consoleplayer then CV_StealthSet(cv_enable, 0) end
		return
	elseif arg == "1" or arg == "on" or arg == "default" or arg == "true" then
		player.saltyhop = true
		if player == consoleplayer then CV_StealthSet(cv_enable, 1) end
		return
	end

	CONS_Printf(player, "saltyhop [on/off] - Toggles whether or not you jump.")
	CONS_Printf(player, "Note that this jump is 100% visual and does not affect gameplay in any way. No jumping over walls or offroad here.")
end)

local lastvalue = cv_enable.value

//It feels wrong using ThinkFrame instead of PlayerThink now
addHook("PostThinkFrame", function()
	-- have to check in thinkframe hook instead of using a cvar function because, and I quote,
	-- "HUD rendering code should not call this function!"
	-- config... oh, you poor thing...
	if lastvalue ~= cv_enable.value then
		COM_BufInsertText(consoleplayer, "saltyhop "..cv_enable.value)
		lastvalue = cv_enable.value
	end

	for player in players.iterate do
		if player.spectator or (player.pflags & PF_TIMEOVER) or player.playerstate == PST_DEAD
		or (cv_hardcode and cv_hardcode.value) then
			local oldspec = player.spectator
			player.spectator = false
			if player.mo.saltyhopclone then
				P_RemoveMobj(player.mo.saltyhopclone)
				player.mo.saltyhopclone = nil
			end
			player.spectator = oldspec

			player.saltyhopstuff = nil
			continue
		end

		local sh = player.saltyhopstuff
		local pmo = player.mo
		local clone = pmo.saltyhopclone

		if not sh then
			sh = {}
			sh.zoffset = 0
			sh.momz = 0
			sh.drift = 0
			player.saltyhopstuff = sh
		end

		if not clone then
			clone = P_SpawnMobj(pmo.x, pmo.y, pmo.z, MT_THOK)
			clone.flags = $ | MF_NOTHINK
			pmo.saltyhopclone = clone
		end

		P_MoveOrigin(clone, pmo.x, pmo.y, pmo.z)

		//Jump Detection
		if player.kartstuff[k_jmp] then
			sh.drift = $ + 1
		else
			sh.drift = 0
		end

		-- landing sound delayed by 1 tic
		if sh.momz == -1 then
			if cv_sfx.value then S_StartSound(clone, sfx_s268) end
			sh.momz = 0
		end

		-- start a hop
		if sh.drift == 1 and P_IsObjectOnGround(pmo)
		and sh.zoffset == 0
		and not player.kartstuff[k_spinouttimer]
		and not player.kartstuff[k_squishedtimer]
		and player.saltyhop and not cv_hide.value then//Yay for customization
			sh.zoffset = -1 -- negative, else the hop ends 1 tic too late
			sh.momz = 27*mapobjectscale/4
			clone.flags2 = $ & ~MF2_DONTDRAW
			if cv_sfx.value then
				S_StartSound(clone, sfx_s25a)
			end
		end

		//Fake jumping
		//It's entirely visual and you're technically still on the ground
		if sh.zoffset ~= 0 then
			//Set all the relative variables
			clone.skin = pmo.skin
			clone.state = pmo.state
			clone.sprite = pmo.sprite
			clone.frame = pmo.frame
			clone.angle = player.frameangle
			clone.color = pmo.color
			clone.colorized = pmo.colorized
			clone.scale = pmo.scale
			clone.height = pmo.height
			clone.flags2 = pmo.flags2 & (MF2_OBJECTFLIP|MF2_SHADOW|MF2_DONTDRAW)
			clone.eflags = pmo.eflags & (MFE_VERTICALFLIP|MFE_DRAWONLYFORP1|MFE_DRAWONLYFORP2|MFE_DRAWONLYFORP3|MFE_DRAWONLYFORP4)

			if pmo.eflags & MFE_JUSTHITFLOOR or player.kartstuff[k_squishedtimer] then
				sh.zoffset = 0
			elseif P_IsObjectOnGround(pmo) then
				sh.zoffset = $ + sh.momz
				sh.momz = $ - (mapobjectscale*3/2)
			else -- in air
				sh.zoffset = $*49/50
				sh.momz = 0
			end

			-- landing?
			if sh.zoffset <= 0 then
				sh.zoffset = 0
				sh.momz = not ((pmo.eflags & MFE_JUSTHITFLOOR) and P_IsObjectOnGround(pmo)) and -1 or 0
			end

			clone.z = $ + sh.zoffset * P_MobjFlip(pmo)
			pmo.flags2 = $ | MF2_DONTDRAW

			if S_SoundPlaying(pmo, sfx_screec)
				S_StopSoundByID(pmo, sfx_screec)
			end
			if S_SoundPlaying(pmo, sfx_drift)
				S_StopSoundByID(pmo, sfx_drift)
			end
		else
			clone.flags2 = $ | MF2_DONTDRAW
		end
	end
end)

addHook("PlayerQuit", function(player)
	if player.mo and player.mo.saltyhopclone then
		P_RemoveMobj(player.mo.saltyhopclone)
	end
end)
