-- CGB HUD refactor
-- Originally by Darkybenji#0648
-- Refactored out of boredness by GenericHeroGuy after the "4 digits crash" bug
-- Now works perfectly with "HUD Transparency" option in the HUD options as well

--Modified to Allow usage of vanilla style speedometor via toggle and colorized by NepDisk

-- G: Oh boy, a double refactor! This was the first script I ever refactored, so tbh it was pretty stinky lol

local speedometers = {
	[1] = { -- kilometers
		symbol = "K_MKMH",
		suffix = " KM/H",
		bg = "K_MBHCGB",
		bgnc = "K_MBNCGB",
		algo = function(p) return FixedDiv(FixedMul(p.speed, 142371), mapobjectscale) / FRACUNIT end,
	},
	[2] = { -- miles
		symbol = "K_MMPH",
		suffix = " MPH",
		bg = "K_MBHCGB",
		bgnc = "K_MBNCGB",
		algo = function(p) return FixedDiv(FixedMul(p.speed, 88465), mapobjectscale) / FRACUNIT end,
	},
	[3] = { -- fracunits
		symbol = "K_MFUT",
		suffix = " FU/T",
		bg = "K_MBHCGB",
		bgnc = "K_MBNCGB",
		algo = function(p) return FixedDiv(p.speed, mapobjectscale) / FRACUNIT end,
	},
	[4] = { -- percent
		symbol = "K_MPER",
		suffix = " %",
		bg = "K_SBHCGB",
		bgnc = "K_SBNCGB",
		algo = function(p) return (FixedDiv(p.speed, FixedMul(K_GetKartSpeed(p, false), 62914)) * 100) >> FRACBITS end,
	}
}

local buttons = {
	{ mask = BT_ACCELERATE, label = "A" },
	{ mask = BT_BRAKE, label = "B", },
	{ mask = BT_DRIFT, label = "D", },
	{ mask = BT_ATTACK, label = "I", },
}

-- stat colors
local speedcolor = SKINCOLOR_BLUEBERRY
local weightcolor = SKINCOLOR_BURGUNDY
local restatcolor = SKINCOLOR_MUSTARD

-- display character stats and input on the bottom left
local cgb_hud = CONFIG_RegisterVar{
	name = "cgb_hud",
	defaultvalue = "On",
	flags = CV_SHOWMODIF,
	possiblevalue = CV_OnOff,
	config_menu = "HUD",
	displayname = "CGB HUD",
	description = "Enable CGB HUD.",
}

-- speedometer style
local cgb_speedometer = CONFIG_RegisterVar{
	name = "cgb_speedometer",
	defaultvalue = "On",
	flags = CV_SHOWMODIF,
	possiblevalue = CV_OnOff,
	config_menu = "HUD",
	displayname = "CGB Speedometer",
	description = "Show the CGB-style speedometer.",
}

-- colorize buttons
local cgb_colorize = CONFIG_RegisterVar{
	name = "cgb_colorize",
	defaultvalue = "On",
	flags = CV_SHOWMODIF,
	possiblevalue = CV_OnOff,
	config_menu = "HUD",
	displayname = "Colorize Input",
	description = "Tint the CGB HUD buttons with your color.",
}

local cgb_namecolor = CONFIG_RegisterVar{
    name = "cgb_skinname_color",
    defaultvalue = "PrefColor",
    possiblevalue = {
        PrefColor = 0,
        PlayerColor = 1,
    },
    config_menu = "HUD",
    displayname = "Skin Name Color",
    description = "Colorize skin name in CGB HUD."
}

-- colorize ribbons
local cgb_colorizebar = CONFIG_RegisterVar{
	name = "cgb_colorizebar",
	defaultvalue = "Off",
	flags = CV_SHOWMODIF,
	possiblevalue = CV_OnOff,
	config_menu = "HUD",
	displayname = "Colorize Ribbons",
	description = "Colorize the CGB HUD ribbons. Inverted if colorizedhud is enabled.",
}

-- Configurable offsets
local cgb_hud_xoffset = CONFIG_RegisterVar{
	name = "cgb_hud_xoffset",
	defaultvalue = "0",
	config_menu = "HUD",
	displayname = "CGB HUD offset",
	description = "Change the offsets for CGB HUD.",
	config_hudmove = "cgb_hud_yoffset",
}

local cgb_hud_yoffset = CV_RegisterVar({
	name = "cgb_hud_yoffset",
	defaultvalue = "0",
})

local cgb_highresportrait = CONFIG_RegisterVar{
	name = "cgb_highresportrait",
	defaultvalue = "Off",
	flags = CV_SHOWMODIF,
	possiblevalue = CV_OnOff,
	config_menu = "HUD",
	displayname = "High Resulution Portrait",
	description = "Use high resolution portrait in CGB HUD. Inverted if highresportrait is enabled.",
}

local cv_highresportrait
local cv_showstats

local function useHighresPortrait()
	if cv_highresportrait then
		return cv_highresportrait.value ~= cgb_highresportrait.value
	end

	return cgb_highresportrait.value
end

local function lookupCvars()
	if cv_highresportrait then return end
	cv_highresportrait = CV_FindVar("highresportrait") or { value = 0 }
	cv_showstats = CV_FindVar("showstats") or { value = 0 }
end

addHook("MapLoad", lookupCvars)
addHook("NetVars", lookupCvars)

local cv_kartdisplayspeed

hud.add(function(v, p, c)
	-- bail if disabled, player doesn't exist, or hardcode stats display is enabled
	if not (cgb_hud.value and p.mo and not cv_showstats.value) then
		hud.enable("speedometer")
		return
	end

	hud.disable("speedometer")

	-- cache cvars
	if cv_kartdisplayspeed == nil then
		cv_kartdisplayspeed = CV_FindVar("kartdisplayspeed") or false
	end

	if SG_HideHud or SG_BattleFullscreen(p) then return end

	local splitindex = c.pnum - 1
	local vflags = SG_GetSplitscreenFlags(splitindex) | V_HUDTRANS | V_SNAPTOLEFT

	local fx = cgb_hud_xoffset.value
	local fy = cgb_hud_yoffset.value + (G_BattleGametype() and (p.kartstuff[k_bumper] and -5 or -8) or 0)
	if ringsOn then
		if p.hasV2Speedometer == 1 then
			fy = $ - 15
		end
	end

	-- 1P or 2P
	if splitscreen <= 1 then
		//Colorized Hardcode support
		local playericoncolor = v.getColormap(p.mo.skin, p.skincolor)
		local inputcolor = cgb_colorize.value and playericoncolor or v.getColormap(p.mo.skin, SKINCOLOR_NICKEL)

		-- colorizedhud flips the meaning of colorizebar
		local colorbg = (cgb_colorizebar.value == 1) ~= v.useColorHud()
		local cmap
		if colorbg then
			cmap = v.getColormap(p.mo.skin, v.getHudColor())
		else
			cmap = v.getColormap(p.mo.skin, SKINCOLOR_NICKEL)
		end

		-- 1P`
		if splitscreen == 0 then
			vflags = $ | V_SNAPTOBOTTOM
			fx = $ + 19
			fy = $ + 154

			-- speedometer
			local spd = speedometers[cv_kartdisplayspeed.value]
			if spd then
				if cgb_speedometer.value then
					v.draw(fx-12, fy-7, v.cachePatch(colorbg and spd.bg or spd.bgnc), vflags, cmap)
					SG_DrawPaddedNum(v, fx+18, fy-9, spd.algo(p), 3, "OPPRNK0", vflags)
					v.draw(fx+22, fy-9, v.cachePatch(spd.symbol), vflags)
				else
					v.drawKartString(fx - 10, fy-16, string.format("%3d", spd.algo(p))..spd.suffix, V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_HUDTRANS, "left")
				end
			end

			-- background
			v.draw(fx-12, fy+8, v.cachePatch(colorbg and "K_BBHCGB" or "K_BBNCGB"), vflags, cmap)

			-- draw buttons
			for i, b in ipairs(buttons) do
				if p.cmd.buttons & b.mask then
					v.draw(fx+10 + i*10, fy+6, v.cachePatch("K_BTDCGB"), vflags, inputcolor)
					v.drawString(fx+13 + i*10, fy+7, b.label, vflags, "thin")
				else -- not presssed
					v.draw(fx+10 + i*10, fy+6, v.cachePatch("K_BTUCGB"), vflags, inputcolor)
					v.drawString(fx+13 + i*10, fy+5, b.label, vflags, "thin")
				end
			end

			-- aim forward/backward
			if p.cmd.buttons & BT_FORWARD then
				v.draw(fx+52, fy+3, v.cachePatch("K_AIFCGB"), vflags)
			end
			if p.cmd.buttons & BT_BACKWARD then
				v.draw(fx+52, fy+16, v.cachePatch("K_AIBCGB"), vflags)
			end

		-- 2P
		else -- if splitscreen == 1
			fx = $ + 19
			fy = $ + 118/2

			-- background
			v.draw(fx-12, fy+8, v.cachePatch(colorbg and "K_MBHCGB" or "K_MBNCGB"), vflags, cmap)

			-- draw buttons
			for i, b in ipairs(buttons) do
				if p.cmd.buttons & b.mask then
					v.drawString(fx+14 + i*6, fy+7, b.label, vflags, "thin")
				else -- not presssed
					v.drawString(fx+14 + i*6, fy+6, SG_Color2Chat[cgb_colorize.value and p.skincolor or SKINCOLOR_GREY]..b.label, vflags, "thin")
				end
			end

			-- aim forward/backward
			if p.cmd.buttons & BT_FORWARD then
				v.draw(fx+37, fy+4, v.cachePatch("K_AIFCGB"), vflags)
			end
			if p.cmd.buttons & BT_BACKWARD then
				v.draw(fx+37, fy+15, v.cachePatch("K_AIBCGB"), vflags)
			end
		end

		-- 1P/2P shared stuff
		local inputturn = p.cmd.driftturn/200 * (encoremode and -1 or 1)
		local wheelflip = inputturn < 0 and V_FLIP or 0
		local wheelx = ((fx+70 - splitscreen*18) * FRACUNIT) - (wheelflip and (FRACUNIT/2)+256 or 0)
		local wheely = (fy+18) * FRACUNIT

		-- wheel
		v.drawScaled(wheelx, wheely, FRACUNIT/2, v.cachePatch("K_WHEEL"..abs(inputturn)), vflags|wheelflip, inputcolor)

		-- character name
        local skinnamecolor = (cgb_namecolor.value == 0) and skins[p.mo.skin].prefcolor or p.skincolor
		v.drawString(fx+20, fy, SG_Color2Chat[skinnamecolor]..skins[p.mo.skin].realname, V_ALLOWLOWERCASE|vflags, "small")

		-- icon and stats
		-- change stat colors if hostmod restat is used

		local facegfxname = useHighresPortrait() and skins[p.mo.skin].facewant or skins[p.mo.skin].facerank

		if APPEAR_GetAppearance and APPEAR_GetAppearance(p) ~= "default" then
			facegfxname = useHighresPortrait() and APPEAR_GetWantedGFX(p) or APPEAR_GetRankGFX(p)
		end

		local facegfx = v.cachePatch(facegfxname)
		local scale = useHighresPortrait() and FRACUNIT/2 or FRACUNIT

		v.drawScaled(fx<<FRACBITS, fy<<FRACBITS, scale, facegfx, vflags, playericoncolor)
		SG_DrawPaddedNum(v, fx+4, fy, p.kartspeed, 1, "OPPRNK0", vflags, v.getColormap(TC_RAINBOW, (p.HMRs and restatcolor or speedcolor)))
		SG_DrawPaddedNum(v, fx+18, fy+8, p.kartweight, 1, "OPPRNK0", vflags, v.getColormap(TC_RAINBOW, (p.HMRs and restatcolor or weightcolor)))

	-- 3P or 4P
	else -- if splitscreen >= 2
		-- align right-side screens
		if splitindex & 1 then
			vflags = $ | V_SNAPTORIGHT ^^ V_SNAPTOLEFT -- cancel out the existing SNAPTOLEFT
			fx = -$ + 121
		else
			fx = $ + 4
		end
		fy = 77 + cgb_hud_yoffset.value -- no battle offset

		-- stats
		if p.HMRs then
			v.drawString(fx, fy, "\x82S"..p.kartspeed.." W"..p.kartweight, vflags, "left")
		else
			v.drawString(fx, fy, "\x88S"..p.kartspeed.." \x87W"..p.kartweight, vflags, "left")
		end
	end
end)
