-- I wanna take a nap
-- splitscreen is unsupported as of yet
-- GenericHeroGuy: now it is! ...sort of
-- G from future: it definitely is now!

local driftrainbow = {
	0, 31, 47, 63, 79, 95, 111, 119, 127, 143, 159, 175, 183, 191, 199, 207, 223, 247
}

local cv_driftgauge = CONFIG_RegisterVar{
	name = "driftgauge",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
	config_menu = "HUD",
	displayname = "Driftgauge",
	description = "Display your drift level below your character."
}

local cv_driftgaugeofs = CONFIG_RegisterVar{
	name = "driftgaugeoffset",
	defaultvalue = "-20",
	flags = CV_FLOAT,
	possiblevalue = { MIN = -FRACUNIT*128, MAX = FRACUNIT*128 },
	config_menu = "HUD",
	displayname = "Driftgauge offset",
	description = "Vertical offset for driftgauge.",
	config_hudmove = true,
	config_hudcall = function(v, p, c, var)
		if not (p.mo and c.chase and p.playerstate == PST_LIVE) then return end
		local basex, basey = SG_GetScreenCoords(v, p, c, p.mo, FixedMul(var.value, var.value > 0 and p.mo.scale or mapobjectscale))
		if not basex then return end
		basex, basey = $>>FRACBITS, $>>FRACBITS
		v.draw(basex, basey, v.cachePatch("K_DGAU"), V_NOSCALESTART|V_OFFSET)
		local dup = v.dupx()
		v.drawFill(basex - dup*23, basey - dup, dup*46, dup*4, (leveltime % 32) | V_NOSCALESTART)
	end
}

local cv_driftgaugetrans = CONFIG_RegisterVar{
	name = "driftgaugetransparency",
	defaultvalue = "Off",
	possiblevalue = CV_OnOff,
	config_menu = "HUD",
	displayname = "Driftgauge transparency",
	description = "Make the driftgauge follow your HUD transparency."
}

local cv_driftgaugecolorized = CONFIG_RegisterVar{
	name = "driftgaugecolorized",
	defaultvalue = "Off",
	possiblevalue = CV_OnOff,
	config_menu = "HUD",
	displayname = "Driftgauge color",
	description = "Colorize driftgauge background. Inverted if colorizedhud is enabled."
}

local cv_driftgaugestyle = CONFIG_RegisterVar {
	name = "driftgaugestyle",
	defaultvalue = "New",
	possiblevalue = { Classic = 0, New = 1, },
	config_menu = "HUD",
	displayname = "Driftgauge style",
	description = "Select driftgauge sprite style."
}

local style = {
    [0] = {
        patch = "K_DGAU",
        cpatch = "K_DGAUC",

        bar = { x = 23, y = 2, width = 47, rows = 4, rowofs = { 0, 0, 0, 0 } },
        num = { x = 32, y = 0, font = "PINGN%d", charwidth = 4 },

        driftskins = {
	        SKINCOLOR_NONE,
	        SKINCOLOR_TEAL,
	        SKINCOLOR_SALMON,
        },

        driftcolors = {
            -- top  --  bottom --
            {0, 0, 10, 16},       -- no drift
            {215, 215, 204, 253}, -- blue
            {125, 125, 151, 159}  -- red
        },
    },

    [1] = {
        patch = "K_NDGAU",
        cpatch = "K_NDGAUC",

        bar = { x = 22, y = -1, width = 46, rows = 3, rowofs = { 0, 1, 0 } },
        num = { x = 22, y = 6, font = "OPPRNK0%d", charwidth = 6 },

        driftskins = {
            [0] = SKINCOLOR_SILVER,
            SKINCOLOR_BLACK,
            SKINCOLOR_BLUE,
            SKINCOLOR_BANANA,
        },

        driftcolors = {
            -- top  --  bottom --
            {9, 1, 9},       -- no drift
            {216, 212, 216}, -- blue
            {126, 124, 126}  -- red
        },
    },
}

local function getDriftgaugeStyle()
    return style[cv_driftgaugestyle.value]
end

local function useColorizedHud(v)
	return v.useColorHud() ~= (cv_driftgaugecolorized.value == 1)
end

local function getBackgroundPatch(v, style)
    return v.cachePatch(style[useColorizedHud(v) and "cpatch" or "patch"])
end

local function getBackgroundColormap(v, p)
    return useColorizedHud(v) and v.getColormap(TC_RAINBOW, v.getHudColor()) or nil
end

-- OPPRNK patches are actually 8 pixels wide rather than the visible 7, bleh
local function drawPaddedNum(v, x, y, num, font, width, flags, cmap)
	for i = #num, 1, -1 do
		local patch = v.cachePatch(string.format(font, string.sub(num, i, i)))
		x = $ - width * (flags & V_NOSCALESTART and v.dupx() or 1)
		v.draw(x, y, patch, flags, cmap)
	end
end

local cv_haste
local function getDriftSparkValue(p)
    if cv_haste == nil then
        cv_haste = CV_FindVar("kmp_haste") or { value = 0 }
    end

    local driftval, driftcharge = K_GetKartDriftSparkValue(p), p.kartstuff[k_driftcharge]

    if cv_haste.value and driftcharge > driftval then -- Ignore haste for blue charges
        local accel = max(0, (9 - p.kartspeed - p.kartweight/2))
		local target = (26*4 + p.kartspeed*2)*8

        local LDC = p.NAlastdriftcharge

        if target < driftval then
            return target, LDC
        end
    end

    return driftval, driftcharge
end

local cv_kartdriftgauge = nil -- Check for hardcode driftgauge
hud.add(function(v, p, c)
    if cv_kartdriftgauge == nil then
        cv_kartdriftgauge = CV_FindVar("kartdriftgauge") or false
    end

    if cv_kartdriftgauge and cv_kartdriftgauge.value then return end

    if replayfreecam then return end

	if not (p.mo and c.chase and p.kartstuff[k_drift] and cv_driftgauge.value and p.playerstate == PST_LIVE) then return end

	local driftval, driftcharge = getDriftSparkValue(p)
	driftcharge = min(driftval*4, $)
	local prerainbow = driftcharge >= driftval*3
	local rainbow = driftcharge >= driftval*4

	local basex, basey = SG_GetScreenCoords(v, p, c, p.mo, FixedMul(cv_driftgaugeofs.value, cv_driftgaugeofs.value > 0 and p.mo.scale or mapobjectscale))
	if not basex then return end
	basex, basey = $>>FRACBITS, $>>FRACBITS
	local dup = v.dupx()

	if v.interpolate then v.interpolate(true) end

	local drifttrans

	if string.find(VERSIONSTRING:lower(), "saturn") and cv_driftgaugetrans.value then -- only use this in saturn since other clients dont support translucent drawfill or stuff will look off!
		drifttrans = v.localTransFlag()
	else
		drifttrans = 0
	end

    local style = getDriftgaugeStyle()

	v.draw(basex, basey, getBackgroundPatch(v, style), V_NOSCALESTART|V_OFFSET|drifttrans, getBackgroundColormap(v, p))

    -- mm no idea how do i customize this via styles table atm
    if cv_driftgaugestyle.value == 1 then
        if prerainbow and not rainbow then
            -- HOT HOT HOT HOT HOOOOOOOT AAAAIIIIIIIIEEEEEEEEEEEEEEEEE
            local trans = abs(sin(leveltime*ANGLE_22h)/(6*FRACUNIT/10))
            v.draw(basex, basey, getBackgroundPatch(v, style), V_NOSCALESTART|V_OFFSET| V_90TRANS - V_10TRANS*trans, v.getColormap(TC_BLINK, SKINCOLOR_RED))
        elseif rainbow then
            local trans = abs(sin(leveltime*ANGLE_22h)/(6*FRACUNIT/20))
            v.draw(basex, basey, getBackgroundPatch(v, style), V_NOSCALESTART|V_OFFSET| V_70TRANS - V_10TRANS*trans, v.getColormap(TC_BLINK, SKINCOLOR_MAROON))
        end
    end

	local barx = basex - dup*style.bar.x
	local bary = basey - dup*style.bar.y
	local BAR_WIDTH = dup*style.bar.width

	local limit = driftval * (driftcharge >= driftval*2 and 2 or 1)
	local width = ((driftcharge - (driftcharge >= driftval and limit or 0)) * BAR_WIDTH) / limit
	local level = min(driftcharge / driftval, 2) + 1

	local cmap
	if driftcharge >= driftval*4 then -- rainbow sparks
		cmap = v.getColormap(TC_RAINBOW, 1 + leveltime % (MAXSKINCOLORS-1))
		for i = 1, style.bar.rows do
			local ofs = style.bar.rowofs[i]*dup
			v.drawFill(barx + ofs, bary+dup*i, BAR_WIDTH - ofs*2, dup, (driftrainbow[(leveltime % #driftrainbow) + 1] + i*2) | V_NOSCALESTART|drifttrans)
		end
	else -- none/blue/red
		cmap = v.getColormap(TC_RAINBOW, style.driftskins[level])
		for i = 1, style.bar.rows do
			local ofs = style.bar.rowofs[i]*dup
			if driftcharge >= driftval then
				v.drawFill(barx + ofs, bary+dup*i, BAR_WIDTH - ofs*2, dup, style.driftcolors[level-1][i] | V_NOSCALESTART|drifttrans)
			end
			local dw = max(0, min(width - ofs, BAR_WIDTH - ofs*2))
			if v.interpolate then v.interpolate(level) end
			v.drawFill(barx + ofs, bary+dup*i, dw, dup, style.driftcolors[level][i] | V_NOSCALESTART|drifttrans)
			if v.interpolate then v.interpolate(true) end
		end
	end

	-- right, also draw a cool number
	drawPaddedNum(v, basex + (dup*style.num.x), basey + dup*style.num.y, string.format("%03d", driftcharge*100 / driftval), style.num.font, style.num.charwidth, V_NOSCALESTART|V_OFFSET|drifttrans, cmap)
end)
