//NameTag by DarkyBenji
--Feature inspired from v2, recreated for v1

--Thanks to Tyron for the Colorskin

-- Edited for Sunflower's Garden. Please grab the original to prevent issues:
-- https://mb.srb2.org/addons/nametags.3055/

-- 10/05/2023: Further improvements and now the line and name end together at all times, similar a v2 footage #authentic
-- 11/05/2023: Now also hides the nametag behind walls!
-- 06/06/2023: Rewrote pretty much everything, nametags look better now
-- 08/06/2023: Change how large nametags are drawn, add icon to small nametags

local starttime = 9*TICRATE + (3*TICRATE/4)
local PST_LIVE = PST_LIVE

local nametagactive = CONFIG_RegisterVar{
	name = "nametag",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
	config_menu = "Nametags",
	displayname = "Enabled",
	description = "Display nametags."
}

-- max distance from which nametags are visible
local nametagdist = CONFIG_RegisterVar{
	name = "nametag_dist",
	defaultvalue = "320",
	possiblevalue = {MIN = 80, MAX = 640},
	config_menu = "Nametags",
	displayname = "Max distance",
	description = "Maximum distance that nametags can be seen from."
}

-- show minimap icon next to nametag
local nametagicon = CONFIG_RegisterVar{
	name = "nametag_icon",
	defaultvalue = "Off",
	possiblevalue = CV_OnOff,
	config_menu = "Nametags",
	displayname = "Show icons",
	description = "Show a player's icon on their nametag."
}

-- max number of nametags onscreen
local nametagcap = CONFIG_RegisterVar{
	name = "nametag_cap",
	defaultvalue = "2",
	possiblevalue = {MIN = 1, MAX = #players},
	config_menu = "Nametags",
	displayname = "Max nametags",
	description = "Maximum number of nametags to display simultaneously."
}

-- small nametags
local nametagsmall = CONFIG_RegisterVar{
	name = "nametag_small",
	defaultvalue = "Off",
	possiblevalue = CV_OnOff,
	config_menu = "Nametags",
	displayname = "Small size",
	description = "Use smaller nametags."
}

-- show clan on nametag
local nametagclan = CONFIG_RegisterVar{
	name = "nametag_clan",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
	config_menu = "Nametags",
	displayname = "Show clan",
	description = "Show a player's Hostmod clan on their nametag."
}

-- show score on nametags
local nametagstar = CONFIG_RegisterVar{
	name = "nametag_star",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
	config_menu = "Nametags",
	displayname = "Show score",
	description = "Display a player's score, represented by stars and crowns."
}

-- show HP on nametags
local nametaghpmod = CONFIG_RegisterVar{
	name = "nametag_hp",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
	config_menu = "Nametags",
	displayname = "Show HP",
	description = "Show a player's HPMOD HP on their nametag."
}

-- transparent nametags (only for saturn client)
local nametagtrans = CONFIG_RegisterVar{
	name = "nametag_transparency",
	defaultvalue = "Off",
	possiblevalue = CV_OnOff,
	config_menu = "Nametags",
	displayname = "Nametags transparency",
	description = "Make Nametags follow your HUD transparency."
}

-- debug
local nametagdebug = CV_RegisterVar({
	name = "nametag_debug",
	defaultvalue = "Off",
	flags = CV_NOSHOWHELP|CV_SHOWMODIF,
	possiblevalue = CV_OnOff,
})

local nametagditto = CV_RegisterVar({
	name = "nametag_ditto",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff,
})

local nametagdittolocal = CV_RegisterVar({
	name = "nametag_ditto_local",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
})

local function getName(p)
	-- (in)sanity checks...
	if nametagditto.value and nametagdittolocal.value and p.dittostuff and p.dittostuff.copied_target and p.dittostuff.copied_target.valid then
		return p.dittostuff.copied_target.name
	end

	return p.name
end

local cv_kartnametag = nil -- Check for hardcode nametag
hud.add(function(v, p, c)
    if cv_kartnametag == nil then
        cv_kartnametag = CV_FindVar("kartnametag") or false
    end

    if cv_kartnametag and cv_kartnametag.value then return end

    if replayfreecam then return end

	local dup = v.dupx()

	if not (nametagactive.value and p.mo) or SG_HideHud or SG_BattleFullscreen(p) then return end

	local tagsdisplayed = 0
	local maxdist = (10 * nametagdist.value) * mapobjectscale
	for target in players.iterate do
		-- that's a lotta conditions!
		if not (target.spectator or target.exiting or target.kartstuff[k_hyudorotimer])
		and target.playerstate == PST_LIVE
		and R_PointToDist(target.mo.x, target.mo.y) < maxdist
		and P_CheckSight(p.mo, target.mo)
		and ((target == p and leveltime < 130) -- display your tag before the countdown
		or (target ~= p and leveltime > starttime) -- then display other players' nametags
		or nametagdebug.value) then -- or just display all of them because yay debug
			local basex, basey, flipped = SG_GetScreenCoords(v, p, c, target.mo, target.mo.height)
			if not basex then continue end
			if p.pflags & PF_FLIPCAM then flipped = (p.mo.eflags & MFE_VERTICALFLIP) ^^ (target.mo.eflags & MFE_VERTICALFLIP) end

			-- nametag limit
			tagsdisplayed = $+1
			if tagsdisplayed > nametagcap.value then break end

			if v.interpolate then v.interpolate(#target) end

			local namex = basex>>FRACBITS
			local namey = basey>>FRACBITS

			local vflags

			if string.find(VERSIONSTRING:lower(), "saturn") and nametagtrans.value then -- only use this in saturn since other clients dont support translucent drawfill or stuff will look off!
				vflags = V_NOSCALESTART|v.localTransFlag() -- cant use V_HUDTRANS here or else stuff disappears during countdown
			else
				vflags = V_NOSCALESTART
			end

			-- get color
			local color = target.mo.color
			local tagcolor = SG_Translations[color][7]

			-- get tag text and length
			local tag = SG_Color2Chat[color]..getName(target)
			if nametagclan.value and target.HMclan then
				tag = "\x8C"..target.HMclan.." "..$
			end
			local str = tag:gsub('l', 'I'):gsub('j', ';'):gsub('g', 'u') -- some characters with wrong widths
			local tagwidth = dup*v.stringWidth(str, 0, "thin")

			-- get icon
			local icon = v.cachePatch(skins[target.mo.skin].facemmap)

			-- now it's time to d-d-d-d-draw!
			if not nametagsmall.value then -- normal size
				if nametagicon.value then tagwidth = $ + dup*(icon.width+1) end

				-- line
				if flipped then
					for i = 1, 4 do
						v.drawFill(namex, namey, dup*3, dup*4, 31 | vflags)
						v.drawFill(namex+dup, namey, dup, dup*4, tagcolor | vflags)
						namey = $ + dup*4
						namex = $ + dup
					end
					namey = $ - dup*3
					v.drawFill(namex, namey+dup*2, dup, dup, 31 | vflags) -- a single black pixel
					v.drawFill(namex+dup, namey, tagwidth, dup*3, 31 | vflags)
					v.drawFill(namex+dup, namey+dup, tagwidth - dup, dup, tagcolor | vflags)
					namex = $ + dup*2
				else -- not flipped
					for i = 1, 4 do
						namey = $ - dup*4
						v.drawFill(namex, namey, dup*3, dup*4, 31 | vflags)
						v.drawFill(namex+dup, namey, dup, dup*4, tagcolor | vflags)
						namex = $ + dup
					end
					v.drawFill(namex, namey, dup, dup, 31 | vflags)
					v.drawFill(namex+dup, namey, tagwidth - dup*2, dup*3, 31 | vflags)
					v.drawFill(namex+dup, namey+dup, tagwidth - dup*3, dup, tagcolor | vflags)
				end

				-- icon
				if nametagicon.value then
					v.draw(namex, namey - dup*(icon.height+1), icon, vflags, v.getColormap(target.mo.skin, color))
					namex = $ + dup*(icon.width+1) -- add offset to other stuff
				end

				-- name and clan
				v.drawString(namex, namey - dup*10, tag, V_ALLOWLOWERCASE | vflags, "thin")

				-- hpmod
				if nametaghpmod.value and hpmod and hpmod.running then
					v.draw(namex, namey - dup*20, v.cachePatch("NTHP"), vflags)
					v.drawString(namex + dup*10, namey - dup*19, "\x8D"..(target.hpmod.hp), vflags, "left")
					continue -- no scores in hpmod
				end

				-- scorekeeper
				if nametagstar.value then
					-- crowns
					for i = 400, target.score, 400 do
						v.draw(namex, namey - dup*19, v.cachePatch("NTCRWN"), vflags)
						namex = $ + dup*16
					end
					local stars = "\x82"
					for i = 100, target.score % 400, 100 do
						stars = $.."*"
					end
					v.drawString(namex, namey - dup*19, stars, vflags, "left")
				end
			else -- small size
				if flipped then
					namey = $ + dup*5
				else -- small offset
					namey = $ - dup*3
				end

				-- icon
				if nametagicon.value then
					v.drawScaled(namex<<FRACBITS, (namey - icon.height/2)<<FRACBITS, FRACUNIT/2, icon, vflags, v.getColormap(target.mo.skin, color))
					namex = $ + dup*(1+icon.width/2) -- add offset to other stuff
				end

				-- name and clan
				v.drawString(namex, namey, tag, V_ALLOWLOWERCASE | vflags, "small")

				-- hpmod
				if nametaghpmod.value and hpmod and hpmod.running then
					v.drawString(namex, namey - dup*5, "\x8D"..(target.hpmod.hp).."%", vflags, "small")
					continue
				end

				-- scorekeeper
				if nametagstar.value then
					local stars = "\x82"
					for i = 100, target.score, 100 do
						stars = $.."*"
					end
					v.drawString(namex, namey - dup*5, stars, vflags, "small")
				end
			end
		end
	end
	if nametagdebug.value then
		v.drawString(12, 125, "Nametags shown: "..tagsdisplayed.."/"..nametagcap.value, V_ALLOWLOWERCASE | V_SNAPTOLEFT, "thin")
	end
end)
