local cv_screenshake = CONFIG_RegisterVar {
    name = "screenshake",
    defaultvalue = "Explosions",
    possiblevalue = {
        Off = 0,
        Explosions = 1,
        Full = 2,
    },

    config_menu = "Other",
    displayname = "Screen Shake",
    description = "Shake the screen in certain situations.",
}

local function set(arr)
    local res = {}

    for _, val in ipairs(arr) do
        res[val] = true
    end

    return res
end

local function rescale(value, oldmin, oldmax, newmin, newmax)
    return newmin + FixedMul(newmax-newmin, FixedDiv(value-oldmin, oldmax-oldmin))
end

local SHAKE_RADIUS = 2000

-- points can be anything that has x, y, and z fields (so mobjs or tables are fine)
local function distance(p1, p2)
    return FixedHypot(FixedHypot(p1.x - p2.x, p1.y - p2.y), p1.z - p2.z)
end

-- Grrr...
local function P_IsDisplayPlayer(p)
    if p == consoleplayer then return true end

    for dp in displayplayers.iterate do
        if dp == p then return true end
    end

    return false
end

local function doScreenShake(epicenter, intensity, time)
    local radius = SHAKE_RADIUS*mapobjectscale

    for p in displayplayers.iterate do
        local oldspec = p.spectator
        p.spectator = false
        local pmo = p.mo
        p.spectator = oldspec

        local dist = distance(pmo, epicenter)

        if dist < radius then
            local useintensity = FixedMul(rescale(dist, 0, radius, FRACUNIT, FRACUNIT/8), intensity)

            P_StartQuake(useintensity, time)
        end
    end
end

local SPIN_SHAKE_MOBJS = set {
    MT_ORBINAUT, MT_ORBINAUT_SHIELD,
    MT_JAWZ, MT_JAWZ_DUD, MT_JAWZ_SHIELD,
    MT_PLAYER,
}

local function bananasnipe(inflictor)
    return inflictor and inflictor.type == MT_BANANA and inflictor.health > 1
end

addHook("PlayerSpin", function(p, inflictor, source)
    if cv_screenshake.value ~= 2 then return end
    if not p.mo then return end -- HOW???

    local snipe = bananasnipe(inflictor)

    if not (snipe or P_IsDisplayPlayer(p) or (inflictor and SPIN_SHAKE_MOBJS[inflictor.type])) then return end

    doScreenShake(p.mo, (snipe and 48 or 24) * (inflictor and inflictor.scale or FRACUNIT), snipe and 12 or 8)
end)

addHook("MobjSpawn", function(mo)
    if cv_screenshake.value == 0 then return end

    doScreenShake(mo, 55 * mo.scale, 18)
end, MT_MINEEXPLOSION)

addHook("MobjSpawn", function(mo)
    if cv_screenshake.value == 0 then return end

    doScreenShake(mo, 55 * mo.scale, 18)
end, MT_SPBEXPLOSION)

addHook("MobjSpawn", function(mo)
    if cv_screenshake.value == 0 then return end

    doScreenShake(mo, 55 * mo.scale, 18)
end, MT_FZEROBOOM)

function A_BallhogExplode(actor, var1, var2)
    super(actor, var1, var2)

    if cv_screenshake.value == 0 then return end

    doScreenShake(actor, 32*actor.scale, 12)
end
