local cv_hitvfx = CONFIG_RegisterVar {
    name = "hitvfx",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,

	config_menu = "Hit VFX",
	displayname = "Hit Visuals",
	description = "Spawn extra visuals when player gets hit.",
}

local cv_hitsfx = CONFIG_RegisterVar {
    name = "hitsfx",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,

	config_menu = "Hit VFX",
	displayname = "Hit Sounds",
	description = "Play extra sounds when player gets hit.",
}

freeslot(
    "sfx_dmga1",
    "sfx_dmga2",
    "sfx_dmga3",
    "sfx_dmga4",
    "sfx_dmgb1",
    "sfx_dmgb2",
    "sfx_dmgb3",
    "sfx_dmgb4"
)

for sfx = sfx_dmga1, sfx_dmgb4 do
    sfxinfo[sfx].flags = $ | SF_X4AWAYSOUND
end

local NUMHITSOUNDS = 4

freeslot(
    "SPR_HFX1"
)

local NUMHITSPRITES = 9

freeslot("MT_HFX", "S_HFX")

states[S_HFX] = {
    sprite = SPR_HFX1,
    frame = A,
    tics = -1,
    nextstate = S_NULL,
}

mobjinfo[MT_HFX] = {
    spawnstate = S_HFX,
    flags = MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT,
}

local NUMFRAMES = {
    [0] = 1,
    [1] = 2,
    [2] = 3,
    [3] = 4,
    [4] = 5,
    [5] = 6,
    [6] = 8,
    [7] = 9,
    [8] = 10,
}

-- I hate freeslot limits
-- There used to be 9 sprites but now they are packed into 1, so i need way to find first frame of a
-- packed sprite
local function pickFrame(num)
    local res = 0

    for i = 0, num-1 do
        res = res + NUMFRAMES[i]
    end

    return res
end

local function lerp(a, b, t)
    return a + FixedMul(b-a, t)
end

local function spawnHitMobj(mo, strength)
    local sprite = lerp(0, NUMHITSPRITES-1, strength)
    local frame = pickFrame(sprite)

    local angle = N_RandomRange(1, 90)*ANG1

    for i = 1, 2 do
        local hit = P_SpawnMobj(
            mo.x,
            mo.y,
            mo.z,
            MT_HFX
        )

        hit.tics = NUMFRAMES[sprite]
        hit.sprite = SPR_HFX1
        hit.frame = frame
        hit.angle = angle + ANGLE_90*i
        hit.scale = mapobjectscale+strength

        K_MatchGenericExtraFlags(hit, mo)
    end
end

local function playHitSound(mo, strength)
    local sfx = (N_RandomKey(2) == 1) and sfx_dmga1 or sfx_dmgb1
    sfx = sfx + lerp(0, NUMHITSOUNDS-1, strength)

    S_StartSound(mo, sfx)
end

local function doHitVFX(mo, strength)
    if cv_hitvfx.value == 1 then
        spawnHitMobj(mo, strength)
    end

    if cv_hitsfx.value == 1 then
        playHitSound(mo, strength)
    end
end

-- Trick to not use too many states because freeslot limit
addHook("MobjThinker", function(mo)
    if not mo.valid then return end -- zzz

    mo.frame = ((mo.frame & FF_FRAMEMASK) + 1) | FF_PAPERSPRITE
end, MT_HFX)

local SHIELD_MOBJS = {
    [MT_ORBINAUT_SHIELD] = true,
    [MT_JAWZ_SHIELD] = true,
}

addHook("PlayerSpin", function(player, inflictor, source)
    if not player.mo then return end -- HOW???

    local imomx, imomy, imomz = 0, 0, 0

    if inflictor then
        if SHIELD_MOBJS[inflictor.type] and inflictor.target then
            imomx, imomy, imomz = inflictor.target.momx, inflictor.target.momy, inflictor.target.momz
        else
            imomx, imomy, imomz = inflictor.momx, inflictor.momy, inflictor.momz
        end
    end

    local speed = FixedHypot(FixedHypot(imomx, imomy), imomz)
    local maxspeed = 100*mapobjectscale

    doHitVFX(player.mo, FixedDiv(min(speed, maxspeed), maxspeed))
end)

--[[
COM_AddCommand("testhitvfx", function(p, strength)
    local s = FRACUNIT/NUMHITSPRITES*tonumber(strength or 1)

    doHitVFX(p.mo, s)
end)
--]]
