local cv_keepmusic = CONFIG_RegisterVar {
    name = "keepmapmusic",
    defaultvalue = "Off",
    possiblevalue = CV_OnOff,

	config_menu = "Other",
	displayname = "Keep Map Music",
	description = "Keep map music playing when reloading same map.",
}

local oldmap = -1
local oldencore = false
local keepmusicname = nil
local keepmusicpos = nil

local STARTTIME = 6*TICRATE + (3*TICRATE/4)
local STARTMUSIC = { KSTART = true, ESTART = true, }
local ENDMUSIC = { KRWIN = true, KROK = true, KRLOSE = true, KRFAIL = true, KBWIN = true, KBOK = true, KBLOSE = true, }

-- Force reset from "key" music to "value" music after map restart, basically
local keepmusic_exceptions = {
    CHRSHF = "CHRSHP",
}

-- Allow game changing music from "key" to "value"
-- Damn you chronoshift...
local changemusic_exceptions = {
    CHRSHF = "CHRSHP",
    CHRSHP = "CHRSHF",
}

local function isStartMusic(newname)
    local votetheme = VOTE_GetCurrentTheme()
    local thememusic = votetheme and votetheme.start_music
    local thememusic_encore = votetheme and votetheme.start_music_encore
    return STARTMUSIC[newname] or (thememusic and thememusic:upper() == newname) or (thememusic_encore and thememusic_encore:upper() == newname)
end

local function isWaitMusic(newname)
    local votetheme = VOTE_GetCurrentTheme()
    local thememusic = votetheme and votetheme.wait_music
    return newname == "WAIT2J" or (thememusic and thememusic:upper() == newname)
end

local function isEndMusic(newname)
    -- For some incredible fucking dumb reason "KRWIN" (passed by game) == "KRWIN" is false, so just
    -- using ENDMUSIC[newname] doesn't work. Maybe its null terminator and something after it in
    -- string passed by game, idk
    for k, _ in pairs(ENDMUSIC) do
        if k:find(newname) then
            return true
        end
    end

    local votetheme = VOTE_GetCurrentTheme()
    local thememusic = votetheme and votetheme.ending_music and votetheme.ending_music:upper()

    if thememusic then
        for _, suffix in ipairs { "RW", "RO", "RL", "RF", "BW", "BO", "BL" } do
            if newname == thememusic..suffix then return true end
        end
    end
end

local function getKeepMusicName()
    local musname = S_MusicName():upper()

    if isStartMusic(musname) or isWaitMusic(musname) or isEndMusic(musname) or musname == "CHALNG" or musname ~= mapmusname:upper() then
        return
    end

    if keepmusic_exceptions[musname] then musname = keepmusic_exceptions[musname] end

    if musname == "" then musname = mapheaderinfo[gamemap].musname:upper() end

    return musname
end

local cv_hardcode -- There's hardcode implementation in saturn, if its enabled then no need to run our version, hardcode will do better
local function isActive()
	if cv_hardcode == nil then
		cv_hardcode = CV_FindVar("keepmusic") or { value = 0 }
	end

	return (cv_hardcode.value == 0) and (cv_keepmusic.value == 1) and (not replayplayback)
end

addHook("MapChange", function(newmap)
    local musname = getKeepMusicName()

    keepmusicname = (oldmap == newmap and oldencore == encoremode) and musname or nil
    keepmusicpos = S_GetMusicPosition()

    oldencore = encoremode
    oldmap = newmap
end)


addHook("MusicChange", function(oldname, newname, ...)
    if not isActive() then return end

    oldname = oldname:upper()
    newname = newname:upper()

    if changemusic_exceptions[oldname] == newname then return end
    if isEndMusic(newname) then return end

    -- in case player used `tunes` to change map music
    local mapmusicchange = newname == mapheaderinfo[gamemap].musname:upper()

    local startmusic = isStartMusic(newname)

    if (startmusic or mapmusicchange) and keepmusicname then
        return keepmusicname, nil, true, keepmusicpos
    end
end)

addHook("PreThinkFrame", function()
    if not isActive() then return end
    if not keepmusicname then return end

    -- Game is about to stop music without warning aaaaaa
    if leveltime == STARTTIME then
        keepmusicpos = S_GetMusicPosition()
    elseif leveltime == STARTTIME+1 then
        S_ChangeMusic(keepmusicname, true, nil, 0, keepmusicpos + 28) -- 28 - 1 tic time (approximately)
    end
end)

-- Kinda a hack but it works, it makes MapChange hook think we change map for first time, forcing it
-- to not keep vote/intermission music
local function resetKeepMusic()
    oldmap = -1
end

addHook("IntermissionThinker", resetKeepMusic)
addHook("VoteThinker", resetKeepMusic)
