-- Custom fastrespawn for Sunflower's Garden
-- Inspired by fast respawn from Juicebox as well as fast respawn from DriftNitro
-- Created by Indev, balance thoughts by Sunflower / AnimeSonic

local TICRATE = TICRATE
local FRACUNIT = FRACUNIT

local fastrespawn = CV_RegisterVar({
    name = "fastrespawn",
    defaultvalue = "On",
    flags = CV_NETVAR | CV_CALL,
    possiblevalue = CV_OnOff,

    func = function(fr)
        print("fastrespawn will be "..(fr.value and "enabled" or "disabled").." next race")
    end,
})

-- Is fastrespawn enabled this race?
local fastrespawn_enabled = fastrespawn.value

addHook("NetVars", function(net)
    fastrespawn_enabled = net($)
end)

-- Fast respawn for people in top half
local function fastrespawnMiddle(p)
    if p.playerstate == PST_DEAD then
        p.deadtimer = max($, TICRATE-1)

        if p.deadtimer == TICRATE-1 then
            p.respawnflash = 7
            S_StartSound(nil, sfx_s3kaf, p)
        end
    end

    if p.kartstuff[k_respawn] then
        p.kartstuff[k_respawn] = min($, (TICRATE*3)/4)
    end
end

-- Fast respawn for people in bottom half
local function fastrespawnBottom(p)
    if p.playerstate == PST_DEAD then
        p.deadtimer = max($, TICRATE-1)

        if p.deadtimer == TICRATE-1 then
            p.respawnflash = 7
            S_StartSound(nil, sfx_s3kaf, p)
        end
    end

    if p.kartstuff[k_respawn] then
        p.kartstuff[k_respawn] = min($, TICRATE/3)
    end
end

addHook("MapChange", function()
    fastrespawn_enabled = fastrespawn.value

    --print("fastrespawn has been "..(fastrespawn_enabled and "enabled" or "disabled"))
end)

addHook("PostThinkFrame", function()
    if not fastrespawn_enabled then return end
    if racecountdown == 1 then return end

    for p in players.iterate do
        if p.spectator then continue end
        if hpmod and hpmod.running and p.hpmod.hp == 0 then continue end

        if p.respawnflash then
            p.respawnflash = p.respawnflash - 1
        end

        -- No fast respawn for frontrunner
        if p.kartstuff[k_position] == 1 then
            continue
        elseif not K_IsPlayerLosing(p) then
            fastrespawnMiddle(p)
        else
            fastrespawnBottom(p)
        end
	end
end)

hud.add(function(v, p, c)
	if p.spectator or (p.pflags & PF_TIMEOVER) or not fastrespawn_enabled or not p.respawnflash then return end
	v.fadeScreen(SKINCOLOR_WHITE, min(10, max(p.respawnflash, 0)))
end)
