-- Lua that makes kitchen sinks eliminate players

local cv_killersinks = CV_RegisterVar({
	name = "killersinks",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
	flags = CV_NETVAR,
})

-- show source player's name
local cv_showsource = CV_RegisterVar({
	name = "killersinks_showsource",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
	flags = CV_NETVAR,
})

-- prevent getting hit during flash tics
local cv_flash = CV_RegisterVar({
	name = "killersinks_flash",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,
	flags = CV_NETVAR,
})

local dramaticsink, dramaticplayer, dramaticsource, dramaticview, dramaticpoint, dramatictimer, dramaticleveltime
local starttime = 6*TICRATE + (3*TICRATE/4)

-- e
local function generateHFForPlayer(p)
	if p and p.valid and p.mo and p.mo.valid then
		return {{p.name, p.mo.hf_overridenamecolor or p.skincolor}, {skins[p.mo.skin].facemmap, p.skincolor}, p}
	else
		return {{}, {}}
	end
end

local function getSinked(player, source)
	-- Eeh had to add this so i could add "SINKED" state in hostmod tab
	-- screen. Probably could be used by other addons in future
	player.sinked = true

	-- War map support (maybe i should add a way to check if "war elim"
	-- gamemode is running?)
	player.clock = 0

	-- Elim support
	if Elimination and Elimination() then
		elim_eliminatePlayer(player)
		return -- Elim will do our job for us ^v^
	end

	player.lives = 0
	player.pflags = $ | PF_TIMEOVER

	local pmo = player.mo

	P_DamageMobj(pmo, nil, nil, 10000)

	K_SpawnMineExplosion(pmo, player.skincolor or SKINCOLOR_KETCHUP)

	local boom = P_SpawnMobj(pmo.x, pmo.y, pmo.z, MT_FZEROBOOM)
	boom.scale = pmo.scale
	boom.angle = pmo.angle
	boom.target = pmo
	boom.pref = player

    -- Need to do that manually, hitfeed won't register killer sinks for some reason
    if hitfeed then
        local v = generateHFForPlayer(player)
        local s = {{}, {}}

        if source then -- there is a source
            s = generateHFForPlayer(source)
        end

        HF_SendHitMessage(s, v, "K_HMSNK")
    end
end

local function sinkCollision(pmo, sink)
	-- another hook might kill the mobjs before we do
	if not (pmo.valid and sink.valid) then return end

	-- height check
	if sink.z > pmo.z + pmo.height then return end
	if sink.z + sink.height < pmo.z then return end

	-- don't hit ourselves, if threshold is non-zero
	-- threshold = throw timer, set to 10 when sink is thrown by player
	if sink.target == pmo and sink.threshold > 0 then return end

	-- don't hit flashing players
	if cv_flash.value and pmo.player.powers[pw_flashing] > 0 then return end

	-- save player names
	local name, sname = pmo.player.name
	if sink.target and sink.target.player then
		sname = sink.target.player.name
	end

	-- not that anyone cares
	if G_BattleGametype() and sink.target and sink.target.player then
		P_AddPlayerScore(sink.target.player, 3)
		K_SpawnBattlePoints(sink.target.player, pmo.player, 3)
	end

	if cv_killersinks.value and not (G_BattleGametype() or pmo.player.pflags & PF_TIMEOVER or pmo.player.exiting or (hpmod and hpmod.running)) then
        local source = sink.target and sink.target.player
		P_KillMobj(sink, pmo, pmo)

		-- count players
		local pcount = 0
		for p in players.iterate do
			if not (p.spectator or p.exiting or p.pflags & PF_TIMEOVER) then pcount = $ + 1 end
		end

		-- if there's 2 or less players left, we're going dramatic
		if pcount <= 2 and not dramaticsink then
			dramaticsink = true
			dramaticplayer = pmo.player
            dramaticsource = source
			dramatictimer = 10*TICRATE
			S_StopMusic()

			-- freeze player movement (and animation, hopefully)
			pmo.flags = $ | MF_NOTHINK

			dramaticview = P_SpawnMobj(pmo.x, pmo.y, pmo.z, MT_THOK)
			dramaticview.state = S_INVISIBLE
			dramaticpoint = P_SpawnMobj(pmo.x + pmo.momx, pmo.y + pmo.momy, pmo.z + pmo.momz, MT_THOK)
			dramaticpoint.state = S_INVISIBLE
			dramaticpoint.angle = pmo.angle
			dramaticleveltime = max(0, leveltime - starttime)
			return -- no cecho
		else
			getSinked(pmo.player, source)
		end
	else
		P_DamageMobj(pmo, sink, sink.target, 10000)
		P_KillMobj(sink, pmo, pmo)
	end

	S_StartSound(nil, sfx_bsnipe)

	if consoleplayer then
		if sname and cv_showsource.value then
			COM_BufInsertText(consoleplayer, 'cecho "'..name..' was hit by\\'..sname..'\'s kitchen sink.\\\\\\\\"')
		else
			COM_BufInsertText(consoleplayer, 'cecho "'..name..'\\was hit by a kitchen sink.\\\\\\\\"')
		end
	end
end

-- why hello there elim!
local function cSpd(ang, spd, scale)
	spd = FixedMul(spd, scale)
	return FixedMul( cos(ang), spd ), FixedMul( sin(ang), spd )
end

-- change dramatic cam angle
local function setView(ang)
	local x, y = cSpd(dramaticpoint.angle + ang, 200*FRACUNIT, mapobjectscale)
	dramaticview.angle = dramaticpoint.angle + ang
	P_SetOrigin(dramaticview, dramaticpoint.x - x, dramaticpoint.y - y, dramaticpoint.z + FixedMul(50*FRACUNIT, mapobjectscale))
end

-- dramatic sink stuff
-- there's all sorts of edge cases to account for here, like the sinked player speccing or just leaving the server
addHook("ThinkFrame", function()
	if leveltime == 0 then return end
	if dramaticsink then
		local playerexists = dramaticplayer and dramaticplayer.valid

		-- force spin animation
		if playerexists then dramaticplayer.kartstuff[k_spinouttimer] = 1 end

		for p in players.iterate do
			p.awayviewmobj = dramaticview
			p.awayviewtics = 1
			if not p.exiting then p.realtime = dramaticleveltime end
			p.powers[pw_nocontrol] = INT16_MAX

			if not (Elimination and Elimination()) then
				p.kartstuff[k_cardanimation] = 0 -- don't play FINISH animation
				if dramatictimer == 3*TICRATE then P_RestoreMusic(p) end
			end
		end

		if dramatictimer == 10*TICRATE then
			setView(ANGLE_315)
			S_StartSound(nil, sfx_bsnipe)
		elseif dramatictimer == 9*TICRATE + TICRATE/2 then
			setView(ANGLE_90)
			S_StartSound(nil, sfx_bsnipe)
		elseif dramatictimer == 9*TICRATE then
			setView(ANGLE_180)
			S_StartSound(nil, sfx_bsnipe)
		elseif dramatictimer == 8*TICRATE then
			if playerexists and not (dramaticplayer.spectator or (dramaticplayer.pflags & PF_TIMEOVER)) then
                if dramaticsource and not dramaticsource.valid then dramaticsource = nil end -- They left?

				getSinked(dramaticplayer, dramaticsource)
			end

			-- don't step on elim's toes
			if not (Elimination and Elimination()) then
				local pcount = 0
				local finished = false
				for p in players.iterate do
					if p.exiting then finished = true end
					if not p.spectator then
						pcount = $ + 1
						-- if sinked player leaves and new player joins, the new player's dramaticplayer will be nil
						if not ((playerexists and p == dramaticplayer) or p.exiting or p.pflags & PF_TIMEOVER) then
							p.exiting = 1
						end
					end
				end
				-- countdown doesn't start if there's only 1 player
				if pcount <= 1 then K_SetExitCountdown(dramatictimer) end

				-- don't cheese records, but don't take away someone's record if they already finished
				if mpr and not finished then mpr.setActive(false) end
			end
		end

		dramatictimer = $ - 1
	end
end)

addHook("MobjCollide", function(thing, tmthing)
	if tmthing.type == MT_PLAYER then
		sinkCollision(tmthing, thing)
		return false
	end
end, MT_SINK)

-- also catch sinks moving into players
addHook("MobjMoveCollide", function(tmthing, thing)
	if thing.type == MT_PLAYER then
		sinkCollision(thing, tmthing)
		return false
	end
end, MT_SINK)

-- Remove "sinked" flags
addHook("MapLoad", function()
	dramaticsink = false
	for player in players.iterate do player.sinked = false end
end)

addHook("NetVars", function(sync)
	dramaticsink = sync($)
	dramaticplayer = sync($)
    dramaticsource = sync($)
	dramaticview = sync($)
	dramaticpoint = sync($)
	dramatictimer = sync($)
	dramaticleveltime = sync($)
end)

hud.add(function(v)
	if dramaticsink then
		local timer = dramatictimer + 1 -- HUD hooks run last, so we're off by one
		if timer == 10*TICRATE
		or timer == 9*TICRATE + TICRATE/2
		or timer == 9*TICRATE then
			v.fadeScreen(0, 10)
		end
	end
end)

--[[
COM_AddCommand("sink", function(p)
	P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z + FRACUNIT*16, MT_SINK)
end)

COM_AddCommand("win", function(p)
	p.starpostnum = INT32_MAX
end)
]]
