-- Custom item odds & calculations without all the other stuff xitemlib comes with
-- Created by GenericHeroGuy for Sunflower's Garden
-- Odds last edited in Lua v231 in Sunflower Conga at January 2, 2024

freeslot("sfx_itroln")
local FRACUNIT, FRACBITS, TICRATE, BT_ATTACK = FRACUNIT, FRACBITS, TICRATE, BT_ATTACK

local REALINDIRECT = 20*TICRATE
local REALBGONE = string.match(VERSIONSTRING, "CEP") and 20*TICRATE or 5*TICRATE

local VANILLA_ODDS = 0
local SUNFLOWER_ODDS = 1
local CONGA_ODDS = 2
local CEP_ODDS = 3
local LOWPOP_ODDS = 4

local VANILLA_CALC = 0
local CONGA_CALC = 1
local TWELVE_CALC = 2
local CEP_CALC = 3

local cv_nrenabled = CV_RegisterVar({
	name = "nr_enabled",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local cv_nrdebug = CV_RegisterVar({
	name = "nr_debug",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local cv_nrodds, cv_nrcalc, initialize

cv_nrodds = CV_RegisterVar({
	name = "nr_odds",
	defaultvalue = "Conga",
	flags = CV_NETVAR|CV_CALL|CV_NOINIT,
	possiblevalue = { Vanilla = VANILLA_ODDS, Sunflower = SUNFLOWER_ODDS, Conga = CONGA_ODDS, CEP = CEP_ODDS, LowPop = LOWPOP_ODDS, },

	func = function()
		initialize()
	end
})

cv_nrcalc = CV_RegisterVar({
	name = "nr_calc",
	defaultvalue = "Conga",
	flags = CV_NETVAR|CV_CALL|CV_NOINIT,
	possiblevalue = { Vanilla = VANILLA_CALC, Conga = CONGA_CALC, ["12-capped"] = TWELVE_CALC, CEP = CEP_CALC },

	func = function()
		initialize()
	end
})

local cv_nrfixmashing = CV_RegisterVar({
	name = "nr_fixmashing",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local cv_nrhyubgone = CV_RegisterVar({
	name = "nr_hyubgone",
	defaultvalue = "30",
	flags = CV_NETVAR,
	possiblevalue = CV_Natural
})

local cv_spbrush = CV_RegisterVar({
	name = "nr_spbrush",
	defaultvalue = "All",
	flags = CV_NETVAR,
	possiblevalue = {
		None = 0,
		Second = 1,
		All = 2,
	},
})

-- first, the tables

local curOdds, curCalc

local odds = {
	[VANILLA_ODDS] = {
		numodds = 10,
		-- vanilla race odds        0  1  2  3  4  5  6  7  8  9
		[KITEM_SNEAKER]         = {20, 0, 0, 4, 6, 7, 0, 0, 0, 0 },
		[KITEM_ROCKETSNEAKER]   = { 0, 0, 0, 0, 0, 1, 4, 5, 3, 0 },
		[KITEM_INVINCIBILITY]   = { 0, 0, 0, 0, 0, 1, 4, 6,10, 0 },
		[KITEM_BANANA]          = { 0, 9, 4, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_EGGMAN]          = { 0, 3, 2, 1, 0, 0, 0, 0, 0, 0 },
		[KITEM_ORBINAUT]        = { 0, 7, 6, 4, 2, 0, 0, 0, 0, 0 },
		[KITEM_JAWZ]            = { 0, 0, 3, 2, 1, 1, 0, 0, 0, 0 },
		[KITEM_MINE]            = { 0, 0, 2, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_BALLHOG]         = { 0, 0, 0, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_SPB]             = { 0, 0, 1, 2, 3, 4, 2, 2, 0,20 },
		[KITEM_GROW]            = { 0, 0, 0, 0, 0, 0, 2, 5, 7, 0 },
		[KITEM_SHRINK]          = { 0, 0, 0, 0, 0, 0, 0, 2, 0, 0 },
		[KITEM_THUNDERSHIELD]   = { 0, 1, 2, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_HYUDORO]         = { 0, 0, 0, 0, 1, 2, 1, 0, 0, 0 },
		[KITEM_POGOSPRING]      = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_KITCHENSINK]     = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLESNEAKER]  = { 0, 0, 0, 0, 3, 7, 9, 2, 0, 0 },
		[KRITEM_TRIPLEBANANA]   = { 0, 0, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TENFOLDBANANA]  = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLEORBINAUT] = { 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_QUADORBINAUT]   = { 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
		[KRITEM_DUALJAWZ]       = { 0, 0, 0, 1, 2, 0, 0, 0, 0, 0 },
	},
     -- EVEN THOUGH ITS CALLED SUNFLOWER ODDS, THIS IS CURRENTLY UNUSED
	[SUNFLOWER_ODDS] = { -- from sg's xitemlib v11
		numodds = 10,
		-- sunflower's race odds    0  1  2  3  4  5  6  7  8  9
		[KITEM_SNEAKER]         = {20, 0, 1, 4, 6, 7, 0, 0, 0, 0 },
		[KITEM_ROCKETSNEAKER]   = { 0, 0, 0, 0, 0, 1, 4, 6, 3, 0 },
		[KITEM_INVINCIBILITY]   = { 0, 0, 0, 0, 0, 1, 4, 6, 10,0 },
		[KITEM_BANANA]          = { 0, 9, 3, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_EGGMAN]          = { 0, 2, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KITEM_ORBINAUT]        = { 0, 7, 7, 4, 1, 0, 0, 0, 0, 0 },
		[KITEM_JAWZ]            = { 0, 0, 3, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_MINE]            = { 0, 0, 2, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_BALLHOG]         = { 0, 0, 0, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_SPB]             = { 0, 0, 1, 2, 3, 4, 2, 2, 0, 20},
		[KITEM_GROW]            = { 0, 0, 0, 0, 0, 0, 2, 5, 7, 0 },
		[KITEM_SHRINK]          = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_THUNDERSHIELD]   = { 0, 1, 2, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_HYUDORO]         = { 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
		[KITEM_POGOSPRING]      = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_KITCHENSINK]     = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLESNEAKER]  = { 0, 0, 0, 0, 4, 7, 9, 2, 0, 0 },
		[KRITEM_TRIPLEBANANA]   = { 0, 0, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TENFOLDBANANA]  = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLEORBINAUT] = { 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_QUADORBINAUT]   = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
		[KRITEM_DUALJAWZ]       = { 0, 0, 0, 1, 1, 0, 0, 0, 0, 0 },
	},
	-- ANIME EDIT THIS ONE, SILLY
	[CONGA_ODDS] = { -- from sg's xitemlib v9
		numodds = 10,
		-- sunflower's conga odds   0  1  2  3  4  5  6  7  8  9
		[KITEM_SNEAKER]         = {20, 0, 0, 6, 7, 0, 0, 0, 0, 0 },
		[KITEM_ROCKETSNEAKER]   = { 0, 0, 0, 0, 0, 4, 3, 6, 3, 0 },
		[KITEM_INVINCIBILITY]   = { 0, 0, 0, 0, 0, 0, 4, 6, 10,0 },
		[KITEM_BANANA]          = { 0, 9, 3, 2, 0, 0, 0, 0, 0, 0 },
		[KITEM_EGGMAN]          = { 0, 2, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KITEM_ORBINAUT]        = { 0, 8, 8, 4, 0, 0, 0, 0, 0, 0 },
		[KITEM_JAWZ]            = { 0, 0, 3, 2, 0, 0, 0, 0, 0, 0 },
		[KITEM_MINE]            = { 0, 0, 2, 1, 1, 0, 0, 0, 0, 0 },
		[KITEM_BALLHOG]         = { 0, 0, 0, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_SPB]             = { 0, 0, 1, 2, 3, 5, 2, 0, 0, 20},
		[KITEM_GROW]            = { 0, 0, 0, 0, 0, 0, 2, 5, 7, 0 },
		[KITEM_SHRINK]          = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_THUNDERSHIELD]   = { 0, 1, 1, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_HYUDORO]         = { 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
		[KITEM_POGOSPRING]      = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_KITCHENSINK]     = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLESNEAKER]  = { 0, 0, 0, 0, 12, 12, 4, 0, 0, 0 },
		[KRITEM_TRIPLEBANANA]   = { 0, 0, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TENFOLDBANANA]  = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLEORBINAUT] = { 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_QUADORBINAUT]   = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
		[KRITEM_DUALJAWZ]       = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
	},

	[CEP_ODDS] = { -- from srb2k-moe-cepdev, 2023-10-22
		numodds = 10,
		-- CEP race odds            0  1  2  3  4  5  6  7  8  9
		[KITEM_SNEAKER]         = {20, 0, 1, 4, 6, 7, 0, 0, 0, 0 },
		[KITEM_ROCKETSNEAKER]   = { 0, 0, 0, 0, 0, 1, 4, 6, 3, 0 },
		[KITEM_INVINCIBILITY]   = { 0, 0, 0, 0, 0, 1, 4, 6,10, 0 },
		[KITEM_BANANA]          = { 0, 9, 4, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_EGGMAN]          = { 0, 2, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KITEM_ORBINAUT]        = { 0, 7, 6, 4, 2, 0, 0, 0, 0, 0 },
		[KITEM_JAWZ]            = { 0, 0, 3, 2, 1, 1, 0, 0, 0, 0 },
		[KITEM_MINE]            = { 0, 0, 2, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_BALLHOG]         = { 0, 0, 0, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_SPB]             = { 0, 0, 1, 2, 3, 4, 2, 2, 0,20 },
		[KITEM_GROW]            = { 0, 0, 0, 0, 0, 0, 2, 5, 7, 0 },
		[KITEM_SHRINK]          = { 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 },
		[KITEM_THUNDERSHIELD]   = { 0, 1, 2, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_HYUDORO]         = { 0, 0, 0, 0, 1, 2, 1, 0, 0, 0 },
		[KITEM_POGOSPRING]      = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_KITCHENSINK]     = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLESNEAKER]  = { 0, 0, 0, 0, 3, 7, 9, 2, 0, 0 },
		[KRITEM_TRIPLEBANANA]   = { 0, 0, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TENFOLDBANANA]  = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLEORBINAUT] = { 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_QUADORBINAUT]   = { 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
		[KRITEM_DUALJAWZ]       = { 0, 0, 0, 1, 2, 0, 0, 0, 0, 0 },
	},

	[LOWPOP_ODDS] = {
		numodds = 10,
		-- Lowpop odds             0   1  2  3  4  5  6  7  8  9
		[KITEM_SNEAKER]         = {20, 0, 0, 6, 7, 0, 0, 0, 0, 0 },
		[KITEM_ROCKETSNEAKER]   = { 0, 0, 0, 0, 0, 4, 3, 6, 3, 0 },
		[KITEM_INVINCIBILITY]   = { 0, 0, 0, 0, 0, 0, 4, 6, 10,0 },
		[KITEM_BANANA]          = { 0, 9, 3, 2, 0, 0, 0, 0, 0, 0 },
		[KITEM_EGGMAN]          = { 0, 2, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KITEM_ORBINAUT]        = { 0, 8, 8, 4, 0, 0, 0, 0, 0, 0 },
		[KITEM_JAWZ]            = { 0, 0, 3, 2, 0, 0, 0, 0, 0, 0 },
		[KITEM_MINE]            = { 0, 0, 2, 1, 1, 0, 0, 0, 0, 0 },
		[KITEM_BALLHOG]         = { 0, 0, 0, 2, 1, 0, 0, 0, 0, 0 },
		[KITEM_SPB]             = { 0, 0, 1, 2, 3, 5, 2, 0, 0, 20},
		[KITEM_GROW]            = { 0, 0, 0, 0, 0, 0, 2, 5, 7, 0 },
		[KITEM_SHRINK]          = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_THUNDERSHIELD]   = { 0, 1, 1, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_HYUDORO]         = { 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
		[KITEM_POGOSPRING]      = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KITEM_KITCHENSINK]     = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLESNEAKER]  = { 0, 0, 0, 0,12,12, 4, 0, 0, 0 },
		[KRITEM_TRIPLEBANANA]   = { 0, 0, 1, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_TENFOLDBANANA]  = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
		[KRITEM_TRIPLEORBINAUT] = { 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 },
		[KRITEM_QUADORBINAUT]   = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
		[KRITEM_DUALJAWZ]       = { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 },
	},
}

local battleodds = {
	numodds = 2,
	--                          0  1
	[KITEM_SNEAKER]         = { 2, 1 },
	[KITEM_ROCKETSNEAKER]   = { 0, 0 },
	[KITEM_INVINCIBILITY]   = { 0, 0 },
	[KITEM_BANANA]          = { 1, 0 },
	[KITEM_EGGMAN]          = { 1, 0 },
	[KITEM_ORBINAUT]        = { 8, 0 },
	[KITEM_JAWZ]            = { 8, 1 },
	[KITEM_MINE]            = { 4, 1 },
	[KITEM_BALLHOG]         = { 2, 1 },
	[KITEM_SPB]             = { 0, 0 },
	[KITEM_GROW]            = { 0, 0 },
	[KITEM_SHRINK]          = { 0, 0 },
	[KITEM_THUNDERSHIELD]   = { 0, 0 },
	[KITEM_HYUDORO]         = { 2, 0 },
	[KITEM_POGOSPRING]      = { 2, 0 },
	[KITEM_KITCHENSINK]     = { 0, 0 },
	[KRITEM_TRIPLESNEAKER]  = { 0, 1 },
	[KRITEM_TRIPLEBANANA]   = { 1, 0 },
	[KRITEM_TENFOLDBANANA]  = { 0, 1 },
	[KRITEM_TRIPLEORBINAUT] = { 2, 0 },
	[KRITEM_QUADORBINAUT]   = { 1, 1 },
	[KRITEM_DUALJAWZ]       = { 2, 1 },
}

local function vanillaDist(player, pingame)
	local pdis = 0
	for p in players.iterate do
		if not p.spectator and p.kartstuff[k_position] ~= 0 and p.kartstuff[k_position] < player.kartstuff[k_position] then
			pdis = $ + FixedHypot(FixedHypot(p.mo.x - player.mo.x,
			                                 p.mo.y - player.mo.y),
			                                 p.mo.z - player.mo.z) / mapobjectscale
			                                 * (pingame - p.kartstuff[k_position])
			                                 / max(1, ((pingame - 1) * (pingame + 1) / 3))
		end
	end
	return pdis
end

local function congaDist(player, pingame)
	local pdis = 0
	-- conga line calc, inspired by Sal's rr-item-cruncher branch
	-- G: yoinked from SG's xItemLib
	local playerMoList = {}
	for p in players.iterate do
		if p.spectator then continue end
		playerMoList[p.kartstuff[k_position]] = p.mo
	end
	local toposition = player.kartstuff[k_position] - 1
	while toposition > 0 do
		local from = playerMoList[toposition + 1]
		local to = nil
		while not to and toposition > 0 do -- Ensures that if somehow the list has any gaps we skip over them, and prevent infinite looping
			to = playerMoList[toposition]
			toposition = $-1
		end
		if to and from then -- if somehow 1st isn't in the list, to would end up nil, so we need to catch that
			--local dist = (FixedHypot(FixedHypot(from.x - to.x, from.y - to.y), from.z - to.z))
			--print("from "..skins[from.skin].name.." to "..skins[to.skin].name.." is "..dist)
			pdis = $ + FixedHypot(FixedHypot(from.x/4 - to.x/4, from.y/4 - to.y/4), from.z/4 - to.z/4) -- trying to not overflow FixedHypot with large distances
			--[[ G: not used anywhere in SG's version??? or vanilla xItemLib???
			if toposition+1 == player.kartstuff[k_position] then -- When we check only the player immmediately ahead
				local ahead  = FixedHypot(FixedHypot(from.x/4 - to.x/4, from.y/4 - to.y/4), from.z/4 - to.z/4)*4
				to = playerMoList[toposition+2]
				if to then
					local behind = FixedHypot(FixedHypot(from.x/4 - to.x/4, from.y/4 - to.y/4), from.z/4 - to.z/4)*4
					local closerange = aggressiondistance*mapobjectscale
					from.xitemcloserange = not (ahead > closerange and behind > closerange)
				end
			end --]]
		end
	end
	pdis = ($ / mapobjectscale)*2 -- Scale it. This results in pdis values a bit weaker in smaller games, and a bit stronger in larger games than the original formula
	pdis = ((100 + 8 - min(pingame, 16)) * $) / 100 -- Again, but this time base it on playercount, same form as the following vanilla adjustment, but much weaker since it stacks with it
	return pdis
end

local function cepDist(player, pingame)
	local pdis = 0

	// Sort all of the players ahead of you.
	// Then tally up their distances in a conga line.

	// This will create a much more consistent item
	// distance algorithm than the "spider web" thing
	// that it was doing before.

	// Add yourself to the list.
	// You'll always be the end of the list,
	// so we can also calculate the length here.
	local sortedPlayers = { [ player.kartstuff[k_position] - 1 ] = player }
	local sortLength = player.kartstuff[k_position]

	// Will only need to do this if there's going to be
	// more than yourself in the list.
	if sortLength > 1 then
		// Add all of the other players.
		for p in players.iterate do
			if p.spectator then continue end

			local pos = p.kartstuff[k_position]

			// Invalid position.
			if not (pos <= 0 or pos > #players
			// Tied / behind us.
			// Also handles ourselves, obviously.
			or pos >= player.kartstuff[k_position])
			// Ties are done with port priority, if there are any.
			and not sortedPlayers[ pos - 1 ] then
				sortedPlayers[ pos - 1 ] = p
			end
		end

		// The chance of this list having gaps is improbable,
		// but not impossible. So we need to spend some extra time
		// to prevent the gaps from mattering.
		local firstPlayer, startFrom
		for i = 0, sortLength-2 do
			if sortedPlayers[i] then
				// First valid index in the list found.
				firstPlayer = sortedPlayers[i]

				// Start the next loop after this player.
				startFrom = i + 1
				break
			end
		end

		if firstPlayer and startFrom < sortLength then
			// First index is valid, so we can
			// start comparing the players.

			for i = startFrom, sortLength-1 do
				if sortedPlayers[i] then
					local secondPlayer = sortedPlayers[i]

					// Add the distance to the player behind you.
					// uses PointToDist2 now for accuracy
					local xydist = R_PointToDist2(firstPlayer.mo.x, firstPlayer.mo.y, secondPlayer.mo.x, secondPlayer.mo.y)
					pdis = $ + R_PointToDist2(0, firstPlayer.mo.z, xydist, secondPlayer.mo.z) / FRACUNIT

					// Advance to next index.
					firstPlayer = secondPlayer
				end
			end
		end
	end

	return pdis
end

local function vanillaMultiplier(pdis, pingame, spbrush)
	if franticitems then // Frantic items make the distances between everyone artifically higher, for crazier items
		pdis = (15 * pdis) / 14
	end
	if spbrush then // SPB Rush Mode: It's 2nd place's job to catch-up items and make 1st place's job hell
		pdis = (3 * pdis) / 2
	end

	return ((28 + (8-pingame)) * pdis) / 28
end

-- same as vanilla, but cap at 12 players
local function twelveMultiplier(pdis, pingame, spbrush)
	if franticitems then // Frantic items make the distances between everyone artifically higher, for crazier items
		pdis = (15 * pdis) / 14
	end
	if spbrush then // SPB Rush Mode: It's 2nd place's job to catch-up items and make 1st place's job hell
		pdis = (3 * pdis) / 2
	end

	return ((28 + (8-min(12, pingame))) * pdis) / 28
end

local function K_ItemOddsScale(numPlayers, spbrush)
	// CEP: due to how baseplayer works, 17P+ lobbies will STILL have the disastrous odds of 0.22 prior, if not WORSE
	// let's try adding another condition

	local basePlayer = 8 // The player count we design most of the game around.
	local vanillaMax = 17 // CEP: Maximum players in "vanilla" (non-30P) clients.
	local extPlayer = 24 // CEP: Cap for 17P+ so that odds don't get too muddled.

	local playerCount = spbrush and 2 or numPlayers
	local playerScaling = 0

	// Then, it multiplies it further if the player count isn't equal to basePlayer.
	// This is done to make low player count races more interesting and high player count rates more fair.
	// (If you're in SPB mode and in 2nd place, it acts like it's a 1v1, so the catch-up game is not weakened.)
	if playerCount < basePlayer then
		// Less than basePlayer: increase odds significantly.
		// 2P: x2.5
		playerScaling = (basePlayer - playerCount) * (FRACUNIT / 4)
	elseif playerCount > basePlayer then
		// More than basePlayer: reduce odds slightly.

		// CEP: 17P+ adjustments
		if playerCount < vanillaMax
			// Less than vanillaMax: Use standard calculations.
			// 16P: x0.6
			playerScaling = (basePlayer - playerCount) * (FRACUNIT / 20)
		elseif playerCount > vanillaMax then
			// More than vanillaMax: Increase odds to fit with the increased playercount
			// 24P: x0.6
			// 30P: x0.45
			playerScaling = (basePlayer - min(extPlayer, playerCount)) * (FRACUNIT / 40) // adding a cap here to be sure
		end
	end

	return playerScaling
end

local function K_ScaleItemDistance(distance, numPlayers, spbrush)
	if mapobjectscale ~= FRACUNIT then
		// Bring back to normal scale.
		distance = FixedDiv(distance, mapobjectscale)
	end

	if franticitems then
		// Frantic items pretends everyone's farther apart, for crazier items.
		distance = (15 * distance) / 14
	end

	if numPlayers > 0 then
		// Items get crazier with the fewer players that you have.
		distance = FixedMul(
			distance,
			FRACUNIT + (K_ItemOddsScale(numPlayers, spbrush) / 2) // Odds are scaled more than the distance brackets
		)
	end

	return distance
end

local function vanillaPowerScale(newodds, pingame, spbrush)
	local PLAYERSCALING = 8 - (spbrush and 2 or min(16, pingame)) -- cap at 16 players
	return FixedMul(newodds<<FRACBITS, FRACUNIT + ((PLAYERSCALING << FRACBITS) / 25)) >> FRACBITS
end

local function cepPowerScale(newodds, pingame, spbrush)
	return FixedMul(newodds, FRACUNIT + K_ItemOddsScale(pingame, spbrush))
end

local function vanillaSpbFunc(newodds, pingame, spbrush, first, second)
	local secondist = 0
	if first and second then // calculate 2nd's distance from 1st, for SPB
		secondist = FixedHypot(FixedHypot(first.mo.x - second.mo.x,
		                                  first.mo.y - second.mo.y),
		                                  first.mo.z - second.mo.z) / mapobjectscale
		if franticitems then
			secondist = (15 * secondist) / 14
		end
		secondist = ((28 + (8-pingame)) * secondist) / 28
	end
	if (secondist/curCalc.distvar) < 3 then return 0 end -- not a bug, vanilla can give SPB negative odds
	return newodds * min((secondist/curCalc.distvar)-4, 3)
end

-- cry about it
local function P_AproxDistance(dx, dy)
	dx = abs(dx)
	dy = abs(dy)
	if dx < dy then return dx + dy - (dx>>1) end
	return dx + dy - (dy>>1)
end

local SPBSTARTDIST = (9*(64*20)) / 2
local SPBFORCEDIST = SPBSTARTDIST + (2*(64*20))

local function cepSpbFunc(newodds, pingame, spbrush, first, second)
	local secondToFirst = 0
	if first and second then
		secondToFirst = P_AproxDistance(P_AproxDistance(first.mo.x - second.mo.x,
	                                                    first.mo.y - second.mo.y),
	                                                    first.mo.z - second.mo.z) / FRACUNIT
		secondToFirst = K_ScaleItemDistance($, pingame, spbrush)
	end

	local distFromStart = max(0, secondToFirst - SPBSTARTDIST)
	local distRange = SPBFORCEDIST - SPBSTARTDIST
	local mulMax = 3*FRACUNIT

	-- CEP triggers signed overflow here at extreme distances, which is undefined behavior
	-- compile with -O1 and you suddendly get no SPBs, compile with -O3 and you suddendly get insane SPB odds
	-- the former is what this script does
	local multiplier = (distFromStart * mulMax) / distRange
	multiplier = max(0, min(multiplier, mulMax))

	return FixedMul(newodds, multiplier)
end

local calcs = {
	[VANILLA_CALC] = {
		distfunc = vanillaDist,
		multiplier = vanillaMultiplier,
		powerscale = vanillaPowerScale,
		spbfunc = vanillaSpbFunc,
		distvar = (64*14),
		spbdist = (64*14)*6,
	},

	[CONGA_CALC] = {
		distfunc = congaDist,
		multiplier = vanillaMultiplier,
		powerscale = vanillaPowerScale,
		spbfunc = vanillaSpbFunc,
		distvar = (64*14),
		spbdist = (64*14)*6,
	},

	[TWELVE_CALC] = {
		distfunc = vanillaDist,
		multiplier = twelveMultiplier,
		powerscale = vanillaPowerScale,
		spbfunc = vanillaSpbFunc,
		distvar = (64*14),
		spbdist = (64*14)*6,
	},

	[CEP_CALC] = {
		distfunc = cepDist,
		multiplier = K_ScaleItemDistance,
		powerscale = cepPowerScale,
		spbfunc = cepSpbFunc,
		distvar = (64*20),
		spbdist = SPBFORCEDIST*6, -- 49920
	},
}

local results = {
	[KITEM_SNEAKER]         = { KITEM_SNEAKER, 1 },
	[KITEM_ROCKETSNEAKER]   = { KITEM_ROCKETSNEAKER, 1, poweritem = true, cooldownstart = true },
	[KITEM_INVINCIBILITY]   = { KITEM_INVINCIBILITY, 1, poweritem = true, cooldownstart = true },
	[KITEM_BANANA]          = { KITEM_BANANA, 1 },
	[KITEM_EGGMAN]          = { KITEM_EGGMAN, 1 },
	[KITEM_ORBINAUT]        = { KITEM_ORBINAUT, 1 },
	[KITEM_JAWZ]            = { KITEM_JAWZ, 1, poweritem = true },
	[KITEM_MINE]            = { KITEM_MINE, 1, poweritem = true, cooldownstart = true },
	[KITEM_BALLHOG]         = { KITEM_BALLHOG, 1, poweritem = true },
	[KITEM_SPB]             = { KITEM_SPB, 1, cooldownstart = true },
	[KITEM_GROW]            = { KITEM_GROW, 1, poweritem = true, cooldownstart = true },
	[KITEM_SHRINK]          = { KITEM_SHRINK, 1, poweritem = true, cooldownstart = true },
	[KITEM_THUNDERSHIELD]   = { KITEM_THUNDERSHIELD, 1, poweritem = true, cooldownstart = true },
	[KITEM_HYUDORO]         = { KITEM_HYUDORO, 1, cooldownstart = true },
	[KITEM_POGOSPRING]      = { KITEM_POGOSPRING, 1 },
	[KITEM_KITCHENSINK]     = { KITEM_KITCHENSINK, 1 },
	[KRITEM_TRIPLESNEAKER]  = { KITEM_SNEAKER, 3, poweritem = true },
	[KRITEM_TRIPLEBANANA]   = { KITEM_BANANA, 3, poweritem = true },
	[KRITEM_TENFOLDBANANA]  = { KITEM_BANANA, 10, poweritem = true },
	[KRITEM_TRIPLEORBINAUT] = { KITEM_ORBINAUT, 3, poweritem = true },
	[KRITEM_QUADORBINAUT]   = { KITEM_ORBINAUT, 4, poweritem = true },
	[KRITEM_DUALJAWZ]       = { KITEM_JAWZ, 2, poweritem = true },
}

local itemdebug = {
	[KITEM_SNEAKER]         = {"Sneaker", V_REDMAP, 129},
	[KITEM_ROCKETSNEAKER]   = {"Rocket Sneakers", V_ORANGEMAP, 86},
	[KITEM_INVINCIBILITY]   = {"Invincibility", V_PURPLEMAP, 194},
	[KITEM_BANANA]          = {"Banana", V_YELLOWMAP, 104},
	[KITEM_EGGMAN]          = {"Eggman Monitor", V_GOLDMAP, 114},
	[KITEM_ORBINAUT]        = {"Orbinaut", V_GRAYMAP, 16},
	[KITEM_JAWZ]            = {"Jawz", V_SKYMAP, 212},
	[KITEM_MINE]            = {"Mine", V_BROWNMAP, 39},
	[KITEM_BALLHOG]         = {"Ballhog", V_LAVENDERMAP, 248},
	[KITEM_SPB]             = {"SPB", V_PINKMAP, 124},
	[KITEM_GROW]            = {"Grow", V_BLUEMAP, 228},
	[KITEM_SHRINK]          = {"Shrink", V_PEACHMAP, 70},
	[KITEM_THUNDERSHIELD]   = {"Thunder Shield", V_TEAMAP, 177},
	[KITEM_HYUDORO]         = {"Hyudoro", nil, 0},
	[KITEM_POGOSPRING]      = {"Pogo Spring", V_ORANGEMAP, 86},
	[KITEM_KITCHENSINK]     = {"Kitchen Sink", nil, 0},
	[KRITEM_TRIPLESNEAKER]  = {"Sneaker x3", V_REDMAP, 128},
	[KRITEM_TRIPLEBANANA]   = {"Banana x3", V_YELLOWMAP, 104},
	[KRITEM_TENFOLDBANANA]  = {"Banana x10", V_YELLOWMAP, 106},
	[KRITEM_TRIPLEORBINAUT] = {"Orbinaut x3", V_GRAYMAP, 16},
	[KRITEM_QUADORBINAUT]   = {"Orbinaut x4", V_GRAYMAP, 18},
	[KRITEM_DUALJAWZ]       = {"Jawz x2", V_SKYMAP, 212},
}

-- then the cvars

local loadcvars = true

local cv_sneaker, cv_rocketsneaker, cv_invincibility, cv_banana, cv_eggmanmonitor
local cv_orbinaut, cv_jawz, cv_mine, cv_ballhog, cv_selfpropelledbomb, cv_grow
local cv_shrink, cv_thundershield, cv_hyudoro, cv_pogospring, cv_kitchensink
local cv_triplesneaker, cv_triplebanana, cv_decabanana
local cv_tripleorbinaut, cv_quadorbinaut, cv_dualjawz
local cv_kartdebugitem, cv_kartdebugamount

function initialize()
	curOdds = G_BattleGametype() and battleodds or odds[cv_nrodds.value]
	curCalc = calcs[cv_nrcalc.value]

	-- CEP modifications
	results[KITEM_EGGMAN].poweritem = cv_nrcalc.value == CEP_CALC
	results[KITEM_EGGMAN].cooldownstart = cv_nrcalc.value == CEP_CALC
	results[KITEM_SPB].cooldownstart = cv_nrcalc.value == CEP_CALC

	if loadcvars then
		loadcvars = false
		cv_sneaker = CV_FindVar("sneaker")
		cv_rocketsneaker = CV_FindVar("rocketsneaker")
		cv_invincibility = CV_FindVar("invincibility")
		cv_banana = CV_FindVar("banana")
		cv_eggmanmonitor = CV_FindVar("eggmanmonitor")
		cv_orbinaut = CV_FindVar("orbinaut")
		cv_jawz = CV_FindVar("jawz")
		cv_mine = CV_FindVar("mine")
		cv_ballhog = CV_FindVar("ballhog")
		cv_selfpropelledbomb = CV_FindVar("selfpropelledbomb")
		cv_grow = CV_FindVar("grow")
		cv_shrink = CV_FindVar("shrink")
		cv_thundershield = CV_FindVar("thundershield")
		cv_hyudoro = CV_FindVar("hyudoro")
		cv_pogospring = CV_FindVar("pogospring")
		cv_kitchensink = CV_FindVar("kitchensink")
		cv_triplesneaker = CV_FindVar("triplesneaker")
		cv_triplebanana = CV_FindVar("triplebanana")
		cv_decabanana = CV_FindVar("decabanana")
		cv_tripleorbinaut = CV_FindVar("tripleorbinaut")
		cv_quadorbinaut = CV_FindVar("quadorbinaut")
		cv_dualjawz = CV_FindVar("dualjawz")
		cv_kartdebugitem = CV_FindVar("kartdebugitem")
		cv_kartdebugamount = CV_FindVar("kartdebugamount")
	end
end

addHook("MapChange", function()
	initialize()
end)
addHook("NetVars", function(sync)
	initialize()
end)

-- then the functions

local precooldown, prebgone
local spbrush, pexiting, mashed, dontforcespb
local starttime = 6*TICRATE + (3*TICRATE/4)

local function K_KartGetItemOdds(pos, item)
	-- finduseodds returned nil, can't get any items
	if pos == nil then return 0 end

	local itemenabled = {
		cv_sneaker.value,
		cv_rocketsneaker.value,
		cv_invincibility.value,
		cv_banana.value,
		cv_eggmanmonitor.value,
		cv_orbinaut.value,
		cv_jawz.value,
		cv_mine.value,
		cv_ballhog.value,
		cv_selfpropelledbomb.value,
		cv_grow.value,
		cv_shrink.value,
		cv_thundershield.value,
		cv_hyudoro.value,
		cv_pogospring.value,
		cv_kitchensink.value,
		cv_triplesneaker.value,
		cv_triplebanana.value,
		cv_decabanana.value,
		cv_tripleorbinaut.value,
		cv_quadorbinaut.value,
		cv_dualjawz.value
	}

	local newodds = curOdds[item][pos]

	if not itemenabled[item] then return 0 end

	// Base multiplication to ALL item odds to simulate fractional precision
	newodds = $ * 4

	local pingame, thunderisout = 0, false
	local first, second
	for p in players.iterate do
		if p.spectator then continue end

		if not G_BattleGametype() or p.kartstuff[k_bumper] then pingame = $ + 1 end

		if p.kartstuff[k_itemtype] == KITEM_THUNDERSHIELD then
			thunderisout = true
		end

		if not G_BattleGametype() then
			if p.kartstuff[k_position] == 1 and not first then
				first = p
			end
			if p.kartstuff[k_position] == 2 and not second then
				second = p
			end
		end
	end

	if item == KITEM_SPB then
		if precooldown > 0 or pexiting > 0 then
			return 0
		elseif pos ~= 10 then
			newodds = curCalc.spbfunc($, pingame, spbrush, first, second)
		end
	end
	if item == KITEM_SHRINK and (precooldown > 0 or (pingame-1) <= pexiting) then return 0 end
	if item == KITEM_THUNDERSHIELD and thunderisout then return 0 end
	if item == KITEM_HYUDORO and prebgone > 0 then return 0 end

	if results[item].cooldownstart and leveltime < (30*TICRATE) + starttime then
		return 0
	end

	if results[item].poweritem then
		if franticitems then
			newodds = $ << 1
		end
		newodds = curCalc.powerscale(newodds, pingame, spbrush)
		if mashed > 0 then
			newodds = FixedDiv(newodds<<FRACBITS, FRACUNIT + mashed) >> FRACBITS
		end
	end

	return newodds
end

local pdis
local disttable = {}
local pretype = {}

local function K_FindUseodds(player, pingame, bestbumper)
	local oddsvalid = {}
	for pos = 1, curOdds.numodds do
		oddsvalid[pos] = false
		for item = 1, #curOdds do
			if K_KartGetItemOdds(pos, item) > 0 then
				oddsvalid[pos] = true
				break
			end
		end
	end

	-- initial distance
	pdis = curCalc.distfunc(player, pingame)

	if G_BattleGametype() then // Battle Mode
		// 2 is the extreme odds of player-controlled "Karma" items
		return pretype[player] == 1 and 2 or 1
	else
		disttable = {}
		if oddsvalid[2] then for i = 1, 1 do disttable[#disttable + 1] = 2 end end
		if oddsvalid[3] then for i = 1, 1 do disttable[#disttable + 1] = 3 end end
		if oddsvalid[4] then for i = 1, 1 do disttable[#disttable + 1] = 4 end end
		if oddsvalid[5] then for i = 1, 2 do disttable[#disttable + 1] = 5 end end
		if oddsvalid[6] then for i = 1, 2 do disttable[#disttable + 1] = 6 end end
		if oddsvalid[7] then for i = 1, 3 do disttable[#disttable + 1] = 7 end end
		if oddsvalid[8] then for i = 1, 3 do disttable[#disttable + 1] = 8 end end
		if oddsvalid[9] then for i = 1, 1 do disttable[#disttable + 1] = 9 end end

		-- UFO bonus
		if cv_nrcalc.value == CEP_CALC and pretype[player] == 1 then
			pdis = 15*pdis/14
		end

		-- player multiplier
		pdis = curCalc.multiplier(pdis, pingame, spbrush)
		-- dumb hack to make SPB forcing work, because pdis is unsigned on CEP
		if cv_nrcalc.value == CEP_CALC and pdis > 32767 then pdis = 123456 end

		if pingame == 1 and oddsvalid[1] then -- nobody's around
			return 1
		elseif pdis <= 0 then -- frontrunning
			return disttable[1]
		elseif player.kartstuff[k_position] == 2 and pdis > curCalc.spbdist and spbplace == -1 and not precooldown and not dontforcespb and oddsvalid[10] then -- force SPB
			return 10
		elseif pdis > curCalc.distvar * ((12 * #disttable) / 14) then -- at rock bottom
			return disttable[#disttable]
		else
			for i = 1, 12 do
				if pdis <= curCalc.distvar * ((i * #disttable) / 14) then
					return disttable[((i * #disttable) / 14) + 1]
				end
			end
		end
	end
end

local function K_KartGetItemResult(p, item)
	p.kartstuff[k_itemtype], p.kartstuff[k_itemamount] = results[item][1], results[item][2]

	if (item == KITEM_SPB or item == KITEM_SHRINK) then
		K_SetIndirectItemCooldown(20*TICRATE)
		precooldown = 20*TICRATE -- don't roll multiple SPBs in one tic
	elseif item == KITEM_HYUDORO then
		K_SetHyudoroCooldown(cv_nrhyubgone.value*TICRATE)
		prebgone = cv_nrhyubgone.value*TICRATE
	-- hardcode rolled an SPB but we didn't
	elseif item ~= KITEM_SPB and item ~= KITEM_SHRINK and indirectitemcooldown == REALINDIRECT then
		K_SetIndirectItemCooldown(precooldown)
	-- hardcode rolled a hyudoro but we didn't
	elseif item ~= KITEM_HYUDORO and hyubgone == REALBGONE then
		K_SetHyudoroCooldown(prebgone)
	end
end

-- and now the interception stuff

local pingame, bestbumper
local preroulette = {}

addHook("PreThinkFrame", function()
	if not cv_nrenabled.value then return end

	-- save these, hardcode may overwrite after player thinkers
	precooldown, prebgone = indirectitemcooldown, hyubgone

	-- cache these instead of computing them for every player
	pingame, bestbumper, pexiting, dontforcespb = 0, 0, 0, false

	for p in players.iterate do
		preroulette[p] = p.kartstuff[k_itemroulette]
		pretype[p] = p.kartstuff[k_roulettetype]
		if not p.spectator then
			// Gotta check how many players are active at this moment.
			pingame = $ + 1
			if p.exiting then pexiting = $ + 1; dontforcespb = true end
			if p.kartstuff[k_bumper] > bestbumper then bestbumper = p.kartstuff[k_bumper] end

			-- debug helper
			if cv_kartdebugitem.value and p.cmd.buttons & BT_CUSTOM2 and P_CanPickupItem(p, 1) then
				K_KartGetItemResult(p, cv_kartdebugitem.value)
				p.kartstuff[k_itemamount] = cv_kartdebugamount.value
				S_StartSound(nil, sfx_itpick, p)
			end
		end
	end

	// No forced SPB in 1v1s, it has to be randomly rolled
	if pingame <= 2 then dontforcespb = true end
end)

addHook("PlayerThink", function(p)
	if not cv_nrenabled.value then return end

	if preroulette[p] and not p.kartstuff[k_itemroulette] then
		if p.kartstuff[k_eggmanexplode] or cv_kartdebugitem.value then return end

		local realtype = p.kartstuff[k_itemtype]
		local realamount = p.kartstuff[k_itemamount]

		if not realtype then
			print("Oops, now's not the time to intercept "..p.name.."'s items")
			return
		end

		// If the roulette finishes or the player presses BT_ATTACK, stop the roulette and calculate the item.
		// Mashing reduces your chances for the good items
		-- it doesn't, because mashing is always set to a negative value which means it never takes effect

		-- earliest time you can stop the roulette
		local roulettestop = TICRATE + (3*(pingame - p.kartstuff[k_position]))

		mashed = ((p.cmd.buttons & BT_ATTACK) and (preroulette[p] + 1) >= roulettestop and not p.kartstuff[k_eggmanheld])
		and (FixedDiv((preroulette[p]+1)*FRACUNIT, ((TICRATE*3)+roulettestop)*FRACUNIT) - FRACUNIT)
		or 0
		if cv_nrfixmashing.value then mashed = abs($) end

		-- special request from anime
		if p.kartstuff[k_itemblinkmode] ~= 2 then
			if mashed < 0 then
--				S_StartSound(nil, sfx_itrolf, p)
				p.kartstuff[k_itemblinkmode] = 0
			elseif mashed > 0 then
				S_StartSound(nil, sfx_itroln, p)
			end
		end

		-- don't need to check this... probably
		-- some leftover from the mario kart days?
		-- dug through commit history, this check was added several months after the change that stopped you from picking up item boxes while holding something already
		-- so I'm just gonna assume it's unnecessary
		if p.kartstuff[k_itemheld] then
			print("Ok I DO need to check for held items (mash "..mashed..")")
			S_StartSound(nil, sfx_ambint)
		end

		local canspbrush = false

		if cv_spbrush.value == 1 then
			canspbrush = p.kartstuff[k_position] == 2
		elseif cv_spbrush.value == 2 then
			canspbrush = p.kartstuff[k_position] > 1
		end

		spbrush = G_RaceGametype() and spbplace ~= -1 and canspbrush

		// Initializes existing spawnchance values
		local spawnchance = {}
		for i = 1, #curOdds do
			spawnchance[i] = 0
		end

		// Split into another function for a debug function below
		local useodds = K_FindUseodds(p, pingame, bestbumper)

		local totalspawnchance = 0
		for item = 1, #curOdds do
			totalspawnchance = $ + K_KartGetItemOdds(useodds, item)
			spawnchance[item] = totalspawnchance
		end

		// Award the player whatever power is rolled
		if totalspawnchance > 0 then
			totalspawnchance = P_RandomKey(totalspawnchance)
			for item = 1, #curOdds do
				if not (spawnchance[item] <= totalspawnchance) then
					K_KartGetItemResult(p, item)
					break
				end
			end
		else
			p.kartstuff[k_itemtype] = KITEM_SAD
			p.kartstuff[k_itemamount] = 1
		end

		if realtype == KITEM_SAD then
			print("Hardcode sad face, I did nothing wrong :^)")
		elseif p.kartstuff[k_itemtype] == KITEM_SAD then
			print("Lua sad face, I did something wrong :(")
		-- note: you need a modded client that doesn't call RNG for this to actually work
		elseif false then
			print("\x82WARNING:\x80 Vanilla mode roulette produced incorrect results!")
			print("Expected "..itemdebug[realtype][1].." x"..realamount..", got "..itemdebug[p.kartstuff[k_itemtype]][1].." x"..p.kartstuff[k_itemamount])
		end
	end
end)

local itemicons
local cache = true
local hudtime = 0

hud.add(function(v, p, c)
	if cache then
		cache = false
		itemicons = {
			[KITEM_SNEAKER]         = v.cachePatch("K_ISSHOE"),
			[KITEM_ROCKETSNEAKER]   = v.cachePatch("K_ISRSHE"),
			[KITEM_INVINCIBILITY]   = v.cachePatch("K_ISINV1"),
			[KITEM_BANANA]          = v.cachePatch("K_ISBANA"),
			[KITEM_EGGMAN]          = v.cachePatch("K_ISEGGM"),
			[KITEM_ORBINAUT]        = v.cachePatch("K_ISORBN"),
			[KITEM_JAWZ]            = v.cachePatch("K_ISJAWZ"),
			[KITEM_MINE]            = v.cachePatch("K_ISMINE"),
			[KITEM_BALLHOG]         = v.cachePatch("K_ISBHOG"),
			[KITEM_SPB]             = v.cachePatch("K_ISSPB"),
			[KITEM_GROW]            = v.cachePatch("K_ISGROW"),
			[KITEM_SHRINK]          = v.cachePatch("K_ISSHRK"),
			[KITEM_THUNDERSHIELD]   = v.cachePatch("K_ISTHNS"),
			[KITEM_HYUDORO]         = v.cachePatch("K_ISHYUD"),
			[KITEM_POGOSPRING]      = v.cachePatch("K_ISPOGO"),
			[KITEM_KITCHENSINK]     = v.cachePatch("K_ISSINK"),
			[KRITEM_TRIPLESNEAKER]  = v.cachePatch("K_ISSHOE"),
			[KRITEM_TRIPLEBANANA]   = v.cachePatch("K_ISBANA"),
			[KRITEM_TENFOLDBANANA]  = v.cachePatch("K_ISBANA"),
			[KRITEM_TRIPLEORBINAUT] = v.cachePatch("K_ISORBN"),
			[KRITEM_QUADORBINAUT]   = v.cachePatch("K_ISORBN"),
			[KRITEM_DUALJAWZ]       = v.cachePatch("K_ISJAWZ"),
		}
	end

	if not (cv_nrdebug.value and cv_nrenabled.value) or c.pnum ~= 1 or p.spectator then return end
	hudtime = $ + 1

	local roulettestop = TICRATE + (3*(pingame - p.kartstuff[k_position]))
	mashed = (p.cmd.buttons & BT_CUSTOM3) and FixedDiv((preroulette[p]+1)*FRACUNIT, ((TICRATE*3)+roulettestop)*FRACUNIT) - FRACUNIT or 0
	if cv_nrfixmashing.value then mashed = abs($) end
	spbrush = G_RaceGametype() and p.kartstuff[k_position] > 1 and spbplace ~= -1

	local useodds = K_FindUseodds(p, pingame, bestbumper)

	-- traditional debugdistribution icons
	local x, y = -9, 98
	for i = 1, #curOdds do
		local itemodds = K_KartGetItemOdds(useodds, i)
		if itemodds <= 0 then continue end

		v.draw(x, y, itemicons[i], V_HUDTRANS)
		v.drawString(x+11, y+31, itemodds, V_HUDTRANS|(itemdebug[i][2] or 0), "thin")
		if results[i][2] > 1 and (hudtime & 2) == 2 then
			for _ = 1, 3 do
				v.drawString(x+11, y+31, itemodds, V_HUDTRANSHALF|V_GRAYMAP, "thin")
			end
		end

		// Display amount for multi-items
		if results[i][2] > 1 then
			v.drawString(x+41, y+31, "x"..results[i][2], V_ALLOWLOWERCASE|V_HUDTRANS, "right")
		end

		x = x + 32
		if x >= 297 then
			x = -9
			y = y + 32
		end
	end

	v.drawString(0, 100, "USEODDS "..(useodds and useodds-1 or "???").." PDIS "..pdis..(p.cmd.buttons & BT_CUSTOM3 and " \x85MASHED" or ""), V_HUDTRANS)
	v.drawString(0, 192, "INDIRECT "..indirectitemcooldown.." HYU "..hyubgone, V_HUDTRANS)

	-- cool-looking graph
	if p.cmd.buttons & BT_CUSTOM1 then
		-- item graphs
		for i = 1, curOdds.numodds do
			local div = 0
			for item = 1, #curOdds do
				div = $ + K_KartGetItemOdds(i, item)
			end
			div = 80*FRACUNIT / (div or 1)

			local y = 100
			for item = 1, #curOdds do
				local useodds = (K_KartGetItemOdds(i, item) * div + FRACUNIT/2)/FRACUNIT
				v.drawFill(i*20, y - useodds, 20, useodds, itemdebug[item][3])
				if results[item][2] > 1 then
					for j = 1, 19, 2 do
						v.drawFill((i*20) + j, y - useodds, 1, useodds, itemdebug[item][3] + 4)
					end
				end
				y = $ - useodds
			end
		end

		-- current odds
		if #disttable < 2 then
			v.draw(0, 0, v.cachePatch("BLANKLVL"))
		else
			local disttablediv = {}
			for i = 1, #disttable do
				disttablediv[disttable[i]] = ($ or 0) + 1
			end
			if not disttablediv[1] then table.remove(disttablediv, 1) end
			local pdis = pdis
			for i = 2, useodds-2 do
				if not disttablediv[i] then return end
				pdis = $ - curCalc.distvar * disttablediv[i]
			end
			v.drawFill(20*useodds + (pdis/(curCalc.distvar/18)) / (disttablediv[useodds-1] or 1), 10, 1, 90, (hudtime & 2) and 2 or 6)
		end
	end
end)
