-- Just so it can be detected by other scripts :3
rawset(_G, "projectilewaterrun", true)

local MOBJS = { MT_ORBINAUT, MT_JAWZ, }

local waterfof = (FF_EXISTS|FF_SWIMMABLE)

-- Precalculated
local PROJMAXSPEED = {
    [0] = 2365440, [1] = 2881536, [2] = 3397632,
}

local function rescale(value, oldmin, oldmax, newmin, newmax)
    return newmin + FixedMul(newmax-newmin, FixedDiv(value-oldmin, oldmax-oldmin))
end

-- Returns true if mobj is on top of woter and can waterrun
-- Dumb game doesn't set MFE_UNDERWATER so i gotta do that myself
local function checkWater(mo)
    local flip = mo.eflags & MFE_VERTICALFLIP

    local touchingwater = false

    for rover in mo.subsector.sector.ffloors() do
        if (rover.flags & waterfof) ~= waterfof then continue end
        if (rover.flags & FF_GOOWATER) == FF_GOOWATER then continue end -- Let this do its job

        local top = rover.topheight
        local bottom = rover.bottomheight

        if rover.t_slope then top = P_GetZAt(rover.t_slope, mo.x, mo.y) end
        if rover.b_slope then bottom = P_GetZAt(rover.b_slope, mo.x, mo.y) end

        -- Fully underwater
        if mo.z > bottom and mo.z + mo.height < top then
            mo.underwater = true
            return false -- underwater mobj can't waterride
        elseif (mo.z < top and mo.z + mo.height > top) or (mo.z < bottom and mo.z + mo.height > bottom) then
            touchingwater = true
        end

        if mo.underwater then continue end -- underwater mobjs can't waterride

        local zdiff, zpos

        local slope = false

        if flip then
            slope = rover.b_slope ~= nil
            zdiff = mo.z + mo.height - bottom
            zpos = bottom - mo.height + 8*FRACUNIT
        else
            slope = rover.t_slope ~= nil
            zdiff = top - mo.z
            zpos = top - 8*FRACUNIT
        end

        -- On sloped water fofs, lets just allow projectiles to be more submerged so they can "climb" the slope up
        if zdiff > 0 and zdiff < mo.height/(slope and 1 or 2) then
            return true, zpos
        end
    end

    -- Reset this flag only when we aren't touching water completely
    if not touchingwater then
        mo.underwater = false
    end

    return false
end

local function thinker(mo)
    if mo.health < 1 then return end

    local MINSPEED_VFX = 14*mo.scale -- Just for vfx
    local MINSPEED = 40*mo.scale
    local MAXSPEED = FixedMul(mo.scale, PROJMAXSPEED[gamespeed])

    local flip = mo.eflags & MFE_VERTICALFLIP
    local goingdown = (flip and mo.momz >= 0) or (not flip and mo.momz <= 0)
    local speed = FixedHypot(mo.momx, mo.momy)

    local on_water, zpos = checkWater(mo)

    if on_water and not P_IsObjectOnGround(mo) and goingdown and abs(mo.momz) < 20*FRACUNIT and speed > MINSPEED then
        mo.z = zpos
        mo.momz = 0

        local scale = FixedMul(FixedDiv(speed - MINSPEED_VFX, MAXSPEED - MINSPEED_VFX), mapobjectscale)

        SG_WaterRunEffect(mo, zpos, scale)
    end
end

for _, mt in ipairs(MOBJS) do
    addHook("MobjThinker", thinker, mt)
end
