-- Minimal time player squish timer may get
local cv_squish_mintime = CV_RegisterVar {
    name = "squish_min_time",
    defaultvalue = TICRATE/4,
    possiblevalue = CV_Natural,
    flags = CV_NETVAR,
}

-- Maximum time (its always TICRATE in vanilla)
local cv_squish_maxtime = CV_RegisterVar {
    name = "squish_max_time",
    defaultvalue = TICRATE,
    possiblevalue = CV_Natural,
    flags = CV_NETVAR,
}

-- If current amount of players is lower than that, always force minimal time
local cv_squish_minplayers = CV_RegisterVar {
    name = "squish_min_players",
    defaultvalue = 3,
    possiblevalue = { MIN = 2, MAX = #players },
    flags = CV_NETVAR,
}

local function rescale(value, oldmin, oldmax, newmin, newmax)
    return newmin + FixedMul(FixedDiv(value-oldmin, oldmax-oldmin), newmax-newmin)
end

addHook("PlayerSquish", function(player, inflictor, source)
    if not (inflictor and inflictor.player) then return end

    local pos = player.kartstuff[k_position]
    local maxpos = 1

    for p in players.iterate do
        -- Multiple players can have same position, so we want to get "maximum theoretically possible" position
        if not (p.spectator and p == player) then maxpos = maxpos + 1 end
    end

    -- We don't set k_squishedtimer yet, just remember which value we should set in PlayerThinker

    if maxpos < cv_squish_minplayers.value then -- Too few players, just force min time
        player.setsquish = cv_squish_mintime.value
    else
        player.setsquish = rescale(pos, maxpos, 1, cv_squish_mintime.value, cv_squish_maxtime.value)
    end
end)

addHook("PlayerThink", function(player)
    if player.setsquish then
        player.kartstuff[k_squishedtimer], player.setsquish = player.setsquish, nil
    end
end)
