-- RouletteHud by Indev for Sunflower's Garden
rawset(_G, "roulettehud", {
    enabled = 1,
    hide = false,
    -- Maybe other fields later?
})

local roulettehud = roulettehud

local cv_roulettehud = CONFIG_RegisterVar {
    name = "luaroulettehud",
    defaultvalue = "On",
    PossibleValue = CV_OnOff,
    flags = CV_CALL,

    func = function(cv)
        roulettehud.enabled = cv.value
    end,

    config_menu = "Lua Roulette Hud",
    displayname = "Enabled",
    description = "Toggle Lua Roulette Hud.",
}

-- Here you go, jin :3
local cv_fancy = CONFIG_RegisterVar {
    name = "fancyroulette",
    defaultvalue = "Off",
    possiblevalue = CV_OnOff,

    config_menu = "Lua Roulette Hud",
    displayname = "Fancy Roulette",
    description = "Toggle extra item roulette animation.",
}

local cv_showamount = CONFIG_RegisterVar {
    name = "showitemamount",
    defaultvalue = "NoExceptions",
    possiblevalue = { WithExceptions = 0, NoExceptions = 1, Force = 2, },

    config_menu = "Lua Roulette Hud",
    displayname = "Show Item Amount",
    description = "Switch between displaying item amount except for certain items (vanilla behavior), always displaying amount for more than 1 item, or forcefully displaying amount.",
}

-- Taken from source
local itemtime = 8*TICRATE
local BASEVIDWIDTH = 320
local BASEVIDHEIGHT = 100

-- Cache for patches
local cache = {}

local function getPatch(v, name)
    if not cache[name] then
        cache[name] = v.cachePatch(name)
    end

    return cache[name]
end

local function singleItemGfx(tinygfx, normalgfx)
    return function(_, tiny)
        return tiny and tinygfx or normalgfx
    end
end

local itemgfx = {
    [KITEM_SNEAKER] = function(amount, tiny)
        if tiny or amount == 1 then return K_GetItemPatch(KITEM_SNEAKER, tiny) end

        return "K_ITSHO"..min(amount, 3)
    end,

    [KITEM_ROCKETSNEAKER] = singleItemGfx("K_ISRSHE", "K_ITRSHE"),

    [KITEM_INVINCIBILITY] = function(_, tiny)
        local frame

        if tiny then
            frame = ((leveltime / 3) % 6) + 1
        else
            frame = ((leveltime / 3) % 7) + 1
        end

        return (tiny and "K_ISINV" or "K_ITINV")..frame
    end,

    [KITEM_BANANA] = function(amount, tiny)
        if tiny or amount == 1 then return K_GetItemPatch(KITEM_BANANA, tiny) end

        return "K_ISBAN"..min(amount, 4)
    end,

    [KITEM_EGGMAN] = singleItemGfx("K_ISEGGM", "K_ITEGGM"),

    [KITEM_ORBINAUT] = function(amount, tiny)
        if tiny then return "K_ISORBN" end

        return "K_ITORB"..min(amount, 4)
    end,

    [KITEM_JAWZ] = function(amount, tiny)
        if tiny or amount == 1 then return K_GetItemPatch(KITEM_JAWZ, tiny) end

        return "K_ISJAW"..min(amount, 2)
    end,

    [KITEM_MINE] = singleItemGfx("K_ISMINE", "K_ITMINE"),

    [KITEM_BALLHOG] = singleItemGfx("K_ISBHOG", "K_ITBHOG"),

    [KITEM_SPB] = singleItemGfx("K_ISSPB", "K_ITSPB"),

    [KITEM_GROW] = singleItemGfx("K_ISGROW", "K_ITGROW"),

    [KITEM_SHRINK] = singleItemGfx("K_ISSHRK", "K_ITSHRK"),

    [KITEM_THUNDERSHIELD] = singleItemGfx("K_ISTHNS", "K_ITTHNS"),

    [KITEM_HYUDORO] = singleItemGfx("K_ISHYUD", "K_ITHYUD"),

    [KITEM_POGOSPRING] = singleItemGfx("K_ISPOGO", "K_ITPOGO"),

    [KITEM_KITCHENSINK] = singleItemGfx("K_ISSINK", "K_ITSINK"),

    [KITEM_SAD] = singleItemGfx("K_ISSAD", "K_ITSAD"),
}

-- First key is 'tiny', second key is item type
local numberdisplaymin_table = {
    [false] = {
        [KITEM_ORBINAUT] = 5,
        [KITEM_SNEAKER] = 4,
        [KITEM_BANANA] = 5,
        [KITEM_JAWZ] = 3,
    },
    [true] = {},
}

-- Items that use dark itembox
local dark = {
    [KITEM_INVINCIBILITY] = true,
    [KITEM_SPB] = true,
    [KITEM_THUNDERSHIELD] = true,
}

local function getItemPatch(v, item, amount, tiny)
    if not itemgfx[item] then item = KITEM_SAD end

    return getPatch(v, itemgfx[item](amount, tiny))
end

local function getItemTimerPatch(v, tiny)
    return getPatch(v, tiny and "K_ISIMER" or "K_ITIMER")
end

-- Splitscreen pain bellow
local cv_item_xoffset, cv_item_yoffset
local function getItemPosition(splitnum)
    if cv_item_xoffset == nil then
        cv_item_xoffset = CV_FindVar("hud_item_xoffset") or false
        cv_item_yoffset = CV_FindVar("hud_item_yoffset") or false
    end

    local offx = 0
    local offy = 0

    if cv_item_xoffset then
        offx = cv_item_xoffset.value
        offy = cv_item_yoffset.value
    end

    local x = 5
    local y = 5

    if splitscreen == 1 then
        y = (splitnum == 0) and 5 or (BASEVIDHEIGHT/2 + 55)
    elseif splitscreen > 1 then
        x = ((splitnum % 2) == 0) and 5 or (BASEVIDWIDTH - 32)
        y = (splitnum < 2) and 5 or (BASEVIDHEIGHT/2 + 55)
    end

    return x + offx, y + offy
end

local function getItemSnapFlags(splitnum)
    if splitscreen == 0 then return V_SNAPTOLEFT|V_SNAPTOTOP end

    if splitscreen == 1 then
        local flags = V_SNAPTOLEFT

        if splitnum == 0 then flags = flags | V_SNAPTOTOP end

        return flags
    elseif splitscreen > 1 then
        local flags = 0

        if (splitnum % 2) == 0 then
            flags = flags | V_SNAPTOLEFT
        else
            flags = flags | V_SNAPTORIGHT
        end

        if splitnum < 2 then flags = flags | V_SNAPTOTOP end

        return flags
    end
end
-- End of splitscreen pain

local restoreItemHud = false
hud.add(function(v, p, c)
    if roulettehud.hide then return end

    if not cv_roulettehud.value then
        if restoreItemHud then hud.enable("item") end

        return
    end

    if p.spectator then return end

    local splitnum = c.pnum - 1

    hud.disable("item")
    restoreItemHud = true

    local ks = p.kartstuff

    local item = ks[k_itemtype]
    local amount = ks[k_itemamount]

    local tiny = splitscreen > 1
    local flipamount = splitscreen > 1 and (splitnum % 2 == 1)

    local colormap

    local x, y
    local flags

    if v.getDrawInfo then
        x, y, flags = v.getDrawInfo("item")
    else
        x, y = getItemPosition(splitnum)
        flags = getItemSnapFlags(splitnum)
    end

    flags = $ | V_HUDTRANS

    local itembar, maxl
    local barlength = tiny and 12 or 26

    local numberdisplaymin

    if cv_showamount.value == 0 then
        numberdisplaymin = numberdisplaymin_table[tiny][item] or 2
    elseif cv_showamount.value == 1 then
        numberdisplaymin = 2
    else
        numberdisplaymin = 1
    end

    if ks[k_itemroulette] then
        item = ((ks[k_itemroulette] / 3) % 14) + 1
        amount = 1
        colormap = v.getColormap(TC_RAINBOW, p.skincolor)
    else
        if ks[k_stolentimer] > 0 then
            if leveltime & 2 then
                item = KITEM_HYUDORO
                amount = 1
            else
                item = 0
            end
        elseif ks[k_stealingtimer] > 0 and (leveltime & 2) then
            item = KITEM_HYUDORO
            amount = 1
        elseif ks[k_eggmanexplode] > 1 then
            if leveltime & 1 then
                item = KITEM_EGGMAN
                amount = 1
            else
                item = 0
            end
        elseif ks[k_rocketsneakertimer] > 1 then
            itembar = ks[k_rocketsneakertimer]
            maxl = (itemtime*3) - barlength

            if leveltime & 1 then
                item = KITEM_ROCKETSNEAKER
            else
                item = 0
            end
        elseif ks[k_growshrinktimer] > 0 then
            if ks[k_growcancel] > 0 then
                itembar = ks[k_growcancel]
                maxl = 26
            end

            if leveltime & 1 then
                item = KITEM_GROW
            else
                item = 0
            end
        elseif ks[k_sadtimer] > 0 then
            if leveltime & 2 then
                item = KITEM_SAD
            else
                item = 0
            end
        else
            if item == 0 or amount <= 0 then return end

            if ks[k_itemheld] and not (leveltime & 1) then
                item = 0
            end

            if ks[k_itemblink] and (leveltime & 1) then
                local skincolor = SKINCOLOR_WHITE

                if ks[k_itemblinkmode] == 2 then
                    skincolor = (1 + (leveltime % (MAXSKINCOLORS-1)))
                elseif ks[k_itemblinkmode] == 1 then
                    skincolor = SKINCOLOR_RED
                end

                colormap = v.getColormap(TC_BLINK, skincolor)
            end
        end
    end

    local bgpatch, bgcolormap = v.getColorHudPatch("item", tiny, dark[item])
    local itemmulpatch, itemmulcolormap = v.getColorHudPatch("itemmul", tiny)

    v.draw(x, y, bgpatch, flags, bgcolormap)

    -- Need to use this if twice, since the mul patch should be drawn before item
    if amount >= numberdisplaymin and not ks[k_itemroulette] then
        v.draw(x + (flipamount and 48 or 0), y, itemmulpatch, flags|(flipamount and V_FLIP or 0), itemmulcolormap)
    end

    if item ~= 0 then
        if ks[k_itemroulette] and cv_fancy.value then
            local offset = (ks[k_itemroulette] % 3) - 1
            local tflags = (offset ~= 0) and ((flags & ~V_HUDTRANS) | V_HUDTRANSHALF) or flags

            if p.deadtimer then offset = 0 end

            if v.interpolate then
                v.interpolate(true)

                if not p.deadtimer then
                    v.interpLatch(true)
                    v.draw(0, (tiny and 6 or 10)*leveltime, v.cachePatch("K_TRNULL"))
                end
            end
            v.draw(x, y + (tiny and 6 or 10)*offset, getItemPatch(v, item, amount, tiny), tflags, colormap)
            if v.interpolate then v.interpolate(false) end
        else
            v.draw(x, y, getItemPatch(v, item, amount, tiny), flags, colormap)
        end
    end

    if amount >= numberdisplaymin and not ks[k_itemroulette] then
        if tiny then
            if flipamount then -- reminder that this is for 3/4p's right end of the screen.
                v.drawString(x+2, y+31, string.format("x%d", amount), V_ALLOWLOWERCASE|flags)
            else
                v.drawString(x+24, y+31, string.format("x%d", amount), V_ALLOWLOWERCASE|flags)
            end
        else
            v.draw(x + 28, y + 41, getPatch(v, "K_ITX"), flags)
            v.drawKartString(x + 38, y + 36, tostring(amount), flags)
        end
    end

    if itembar then
        local fill = (itembar * barlength) / maxl
        local length = min(barlength, fill)
        local height = (not tiny) and 2 or 1

        local bx = tiny and 17 or 11
        local by = tiny and 27 or 35

        v.draw(x + bx, y + by, getItemTimerPatch(v, tiny), flags)
        v.drawFill(x + bx + 1, y + by + 1, (length == 2) and 2 or 1, height, 12|flags)

        if length > 2 then
            v.drawFill(x + bx + length, y + by + 1, 1, height, 12|flags)

            if height == 2 then
                v.drawFill(x + bx + 2, y + by + 2, length-2, 1, 8|flags)
            end

            v.drawFill(x + bx + 2, y + by + 1, length-2, 1, 120|flags)
        end
    end

    if ks[k_eggmanexplode] then
        local num = min(ks[k_eggmanexplode]/TICRATE, 3)
        v.draw(x + 17, y + 13 - (tiny and 1 or 0), getPatch(v, "K_EGGN"..num), flags)
    end
end)
