-- Custom votescreen by Indev
-- Cycles between different vote backgrounds (and theme)
-- Only compatible with the Saturn Client :)
-- G: only the backgrounds :yumetroll:

-- Themes do NOT go there
local themes = {}

rawset(_G, "VOTE_AddTheme", function(theme)
	-- no dupes!
	for _, v in pairs(themes) do
		if theme.name == v.name then return end
	end
	print("Adding theme "..theme.name)
	table.insert(themes, theme)
end)

-- Forward declaration...
local gettheme
local cv_preftheme = {}

-- Netvars
local changetheme_tic = false
local changetheme = 0
local themenum = 0
local forcetheme = 0

local cv_changetheme_interval = CV_RegisterVar {
    name = "votetheme_change_interval",
    defaultvalue = "1",
    possiblevalue = CV_Unsigned,
    flags = CV_NETVAR,
}

local votetheme_prefer_cons_t = {Off = 0}

local function setBg()
	local theme = gettheme()
	if theme and hud.setVoteBackground then
		hud.setVoteBackground(G_BattleGametype() and theme.battletexture or theme.texture)
	end
end

local cv_extended = CV_RegisterVar {
    name = "votetheme_music",
    defaultvalue = "On",
    PossibleValue = { Off = 0, On = 1, Force = 2},
}

-- turned into a command because netvars can't be registered *after* joining a netgame
COM_AddCommand("votetheme_force", function(p, choice)
	local themeval, themename
	for k, v in pairs(votetheme_prefer_cons_t) do
		if not choice and forcetheme == v then
			return CONS_Printf(p, "Forced theme is "..k)
		end
		if tonumber(choice) == v or (choice and k:lower() == choice:lower()) then
			themename, themeval = k, v
		end
	end

	if themeval == nil then
		return CONS_Printf(p, "Unknown theme "..choice)
	end

	CONS_Printf(p, "Forcing theme "..themename)
	forcetheme = themeval
	setBg()
end, COM_ADMIN)

-- cvars must be registered late, otherwise you'll end up with no themes to pick from
local registered = false
local function CheckCvars()
	if registered then return end
	registered = true

	for i = 1, #themes do
		local name = themes[i].name or "Theme"..i
		votetheme_prefer_cons_t[name] = i
	end

	cv_preftheme = CV_RegisterVar {
		name = "votetheme_prefer",
		defaultvalue = "Off",
		PossibleValue = votetheme_prefer_cons_t,
		flags = CV_CALL|CV_NOINIT,
		func = setBg
	}
end

gettheme = function()
    local num = themenum

    if forcetheme ~= 0 then
        num = forcetheme
    end

    if cv_preftheme.value ~= 0 then
        num = cv_preftheme.value
    end

    return themes[num]
end

rawset(_G, "VOTE_GetCurrentTheme", gettheme)

addHook("IntermissionThinker", function()
    if changetheme_tic then
        if CV_FindVar("advancemap").value == 3 then
            changetheme = max(0, changetheme-1)
        end
        changetheme_tic = false
    end
    if changetheme == 0 then
        changetheme = cv_changetheme_interval.value

        themenum = P_RandomRange(1, #themes)
		setBg()
    end
end)

local replacements = {
	vote = { extended = false, key = "vote_music" },
	voteea = { extended = false, key = "vote_music", suffix = "a" },
	voteeb = { extended = false, key = "vote_music", suffix = "b" },
	racent = { extended = true, key = "intermission_music" },
	krwin = { extended = true, key = "ending_music", suffix = "rw" },
	krok = { extended = true, key = "ending_music", suffix = "ro" },
	krlose = { extended = true, key = "ending_music", suffix = "rl" },
	krfail = { extended = true, key = "ending_music", suffix = "rf" },
	kbwin = { extended = true, key = "ending_music", suffix = "bw" },
	kbok = { extended = true, key = "ending_music", suffix = "bo" },
	kblose = { extended = true, key = "ending_music", suffix = "bl" },
	kstart = { extended = true, key = "start_music" },
	estart = { extended = true, key = "start_music_encore" },
	wait2j = { extended = true, key = "wait_music" },
}

addHook("MusicChange", function(_, musname)
	CheckCvars()

    -- :Trollic:
    if not (hud.setVoteBackground or cv_extended.value == 2) then return end

    local theme = gettheme()
    local musname = musname:lower()
	if not theme then return end

	local rep = replacements[musname]
	if rep and not (rep.extended and not cv_extended.value) then
		local music = theme[rep.key] and theme[rep.key]..(rep.suffix or "")

		if music and S_MusicExists(music) then return music end
	end
end)

addHook("MapLoad", function()
    changetheme_tic = true
    CheckCvars()
end)

addHook("NetVars", function(net)
    changetheme_tic = net(changetheme_tic)
    changetheme = net(changetheme)
    themenum = net(themenum)
	forcetheme = net(forcetheme)
	setBg() -- otherwise it doesn't show up when joining during intermission
end)
