-- Joinblock by Indev for Sunflower's Garden
-- Disables joining for known midjoin desynch maps to provide a better experience
local MAPS = {
    MAPZK = true, -- Flesh Paradise Zone; moving plats, desynchs in vanilla
    MAPUT = true, -- Bellbridge Zone; moving plats, desynchs in vanilla
    MAPA4 = true, -- Digital Dimension Zone; Gliders
    MAP59 = true, -- Opulence Zone; moving plats, desynchs in vanilla
    MAP36 = true, -- Kodachrome Void Zone; moving plats, desynchs in vanilla
    MAPLS = true, -- Kodachrome Palace; moving plats, desynchs in vanilla
    MAPR9 = true, -- Rusty Rig Zone; Unknown, desynchs often
    MAPWN = true, -- Smash Kart Circuit; Gliders
    MAPXS = true, -- Windmill isle; Gliders
    MAPZH = true, -- Windy Valley; A mess
	MAPX4 = true, -- Windy Canyon; Gliders
    MAPVE = true, -- Ironbed Descent Zone; Gliders
	MAP50 = true, -- MKSC Rainbow Road
    MAPTY = true, -- SRB2k Mushroom Plunge; Gliders
}

local joinblock = false

-- i wish these were a thing like leveltime
local intermissiontime = 0
local votetime = 0

local function countPlayers()
    local c = 0

    for player in players.iterate do
        if not player.spectator then
            c = c + 1
        end
    end

    return c
end

local cv_joinrefusemessage
local function setJoinAllowed(allow)
    if not isserver then return end

    -- On latest saturn, there is separate CV_HIDE variable `allowsaturnjoin`, which is kinda
    -- created for this script. When we disallow joins due to synch reasons, we want to still allow
    -- saturn clients to connect, since they have certain desynch fixes and improved resynching
    local set = CV_StealthSet

    if allow then
        -- Trigger onchange function, so allowsaturnjoin always gets enabled, in case admin toggled
        -- allowjoin off
        --
        -- Probably there should be a check that if allowjoin was off it won't get toggled on by
        -- this script, but eee thats kinda complicated thing to do, for now at least
        set = CV_Set
    end

    set(CV_FindVar("allowjoin"), allow and 1 or 0)

    if not cv_joinrefusemessage then cv_joinrefusemessage = CV_FindVar("joinrefusemessage") end
    if not cv_joinrefusemessage then return end -- Didn't find it

    if allow then
        CV_Set(cv_joinrefusemessage, cv_joinrefusemessage.defaultvalue)
    else
        CV_Set(cv_joinrefusemessage, "Desynch-heavy map is currently in play. Try again in 2 minutes.")
    end
end

local function updateJoinblock(var)
    if var.value then
        setJoinAllowed(countPlayers() < 2 or not joinblock)
    else
        setJoinAllowed(true)
    end
end

local cv_joinblock = CV_RegisterVar {
    name = "joinblock",
    defaultvalue = "On",
    PossibleValue = CV_OnOff,
    flags = CV_NETVAR|CV_CALL|CV_NOINIT,
    func = updateJoinblock,
}

addHook("MapLoad", function(mapid)
    joinblock = MAPS[G_BuildMapName(mapid)]
    intermissiontime = 0
    votetime = 0
end)

addHook("ThinkFrame", function()
    if not isserver then return end

    if leveltime == 1 then
        updateJoinblock(cv_joinblock)
    end

    if joinblock and (countPlayers() < 2) then
        joinblock = false
        updateJoinblock(cv_joinblock)
    end
end)

addHook("IntermissionThinker", function()
    if not isserver then return end

    joinblock = false
    intermissiontime = $ + 1

    if intermissiontime == 1 then
        updateJoinblock(cv_joinblock)
    end
end)

addHook("VoteThinker", function()
    if not isserver then return end

    joinblock = false
    votetime = $ + 1

    if votetime == 1 then
        updateJoinblock(cv_joinblock)
    end
end)
