-- Colorize projectiles (and rocket sneakers) by Indev for Sunflower's Garden
-- Idea by AnimeSonic after being inspired by CGB having colourized Jawz. We always like to go further.

local rainbowcolor = 1

local MAXSKINCOLORS = MAXSKINCOLORS
local k_invincibilitytimer = k_invincibilitytimer
local pw_invulnerability = pw_invulnerability
local MT_BANANA = MT_BANANA

-- Need to have additional behavior
local shield = {
    [MT_JAWZ_SHIELD] = true,
    [MT_ORBINAUT_SHIELD] = true,
    [MT_BANANA_SHIELD] = true,
    [MT_SSMINE] = true,
    [MT_EGGMANITEM_SHIELD] = true,
    [MT_SINK_SHIELD] = true,

    [MT_ROCKETSNEAKER] = true, -- Not really a projectile
}

local nocolorize = {
    [MT_ORBINAUT] = true,
    [MT_ORBINAUT_SHIELD] = true,

    [MT_JAWZ] = true,
    [MT_JAWZ_DUD] = true,
    [MT_JAWZ_SHIELD] = true,

    [MT_SSMINE] = true,
    [MT_SSMINE_SHIELD] = true,

    [MT_ROCKETSNEAKER] = true,

    [MT_BANANA_SHIELD] = true,
}

local projectiles = {
    MT_JAWZ, MT_JAWZ_DUD, MT_JAWZ_SHIELD,
    MT_ORBINAUT, MT_ORBINAUT_SHIELD,
    MT_BANANA, MT_BANANA_SHIELD,
    MT_SSMINE, MT_SSMINE_SHIELD,
    MT_EGGMANITEM, MT_EGGMANITEM_SHIELD,
    MT_SINK, MT_SINK_SHIELD,
    MT_BALLHOG,
    MT_SPB,

    MT_ROCKETSNEAKER, -- Not really a projectile
}

local cvars_to_reg = {
    { name = "jawz", mobjs = { MT_JAWZ, MT_JAWZ_DUD, MT_JAWZ_SHIELD, }, default = "On", },
    { name = "orbinauts", mobjs = { MT_ORBINAUT, MT_ORBINAUT_SHIELD, }, default = "On", },
    { name = "bananas", mobjs = { MT_BANANA, MT_BANANA_SHIELD, }, default = "On", },
    { name = "mines", mobjs = { MT_SSMINE, MT_SSMINE_SHIELD, }, default = "On", },
    { name = "eggmanitems", mobjs = { MT_EGGMANITEM, MT_EGGMANITEM_SHIELD, }, default = "Off", },
    { name = "sinks", mobjs = { MT_SINK, MT_SINK_SHIELD, }, default = "On", },
    { name = "ballhogs", mobjs = { MT_BALLHOG, }, default = "On", },
    { name = "rocketsneakers", mobjs = { MT_ROCKETSNEAKER, }, default = "On", },
    { name = "selfpropelledbombs", mobjs = { MT_SPB, }, default = "On", },
}

local cvars = {}

for _, def in ipairs(cvars_to_reg) do
    local cv = CV_RegisterVar {
        name = "colorize_"..def.name,
        defaultvalue = def.default or "On",
        possiblevalue = CV_OnOff,
        flags = CV_NETVAR, -- I would leave customization to client but doing that is instant synch failure :(
    }

    for _, mt in ipairs(def.mobjs) do
        cvars[mt] = cv
    end
end

-- Mines are bitches
local mines_list = {}

addHook("MobjSpawn", function(mo)
    table.insert(mines_list, mo)
end, MT_SSMINE_SHIELD)

addHook("NetVars", function(net)
    mines_list = net(mines_list)
end)

addHook("MapLoad", function()
    mines_list = {}
end)

local function colorize(mo)
    if not (cvars[mo.type] and cvars[mo.type].value) then
        return
    end

    if (mo.target and mo.target.valid and mo.target.player) and (not mo.colorinit or shield[mo.type]) then
        local player = mo.target.player

        mo.colorinit = true

        mo.rainbow = player.kartstuff[k_invincibilitytimer] > 0 or player.powers[pw_invulnerability] > 0

        mo.target_color = player.skincolor
    end

    -- Special case: landed bananas should not be colorized
    local banana = mo.type == MT_BANANA and mo.health <= 1

    if mo.rainbow or not (nocolorize[mo.type] or banana) then
        mo.colorized = true
    else
        mo.colorized = false
    end

    if mo.rainbow then
        mo.color = rainbowcolor
        K_SpawnSparkleTrail(mo)
    elseif mo.target_color then
        mo.color = mo.target_color
    end
end

for _, mt in ipairs(projectiles) do
    assert(mt) -- Don't want to accidently add colorize thinker to ALL mobj's :D

    addHook("MobjThinker", colorize, mt)
end

addHook("ThinkFrame", function()
    -- Optimization, only calculate this thing once per frame
    rainbowcolor = (1+leveltime%(MAXSKINCOLORS-1))

    for player in players.iterate do
        local pmo = player.mo

        if pmo then
            local cur = pmo.hnext

            while cur do
                if not (cvars[cur.type] and cvars[cur.type].value) then
                    cur = cur.hnext
                    continue
                end

                if cur.rainbow then
                    cur.color = rainbowcolor
                end

                cur = cur.hnext
            end

            for _, mine in ipairs(mines_list) do
                if mine.valid and mine.rainbow then
                    if not (cvars[mine.type] and cvars[mine.type].value) then continue end
                    mine.color = rainbowcolor
                end
            end
        end
    end
end)

-- Not really related to colorized projectiles but its such small change idk where to put it

-- e
local function K_FlipFromObject(mo, master)
	mo.eflags = (mo.eflags & ~MFE_VERTICALFLIP) | (master.eflags & MFE_VERTICALFLIP)
	mo.flags2 = (mo.flags2 & ~MF2_OBJECTFLIP) | (master.flags2 & MF2_OBJECTFLIP)

	if mo.eflags & MFE_VERTICALFLIP then
		mo.z = $ + master.height - FixedMul(master.scale, mo.height)
    end
end

local function spawnPoof(x, y, z, color)
    local radius = 60*mapobjectscale
    for i = 0, 5 do
        local angle = 60*ANG1*i
        local offx = FixedMul(radius, sin(angle))
        local offy = FixedMul(radius, cos(angle))

        local poof = P_SpawnMobj(x + offx, y + offy, z, MT_WIPEOUTTRAIL)
        poof.colorized = true
        poof.color = color
    end
end

local function spawnSparks(x, y, z)
    local radius = 60*mapobjectscale
    for color = (leveltime % 4), MAXSKINCOLORS, 4 do
        local angle = N_RandomRange(0, 355)*ANG1

        local dirx, diry = sin(angle), cos(angle)

        local offx = FixedMul(radius, dirx)
        local offy = FixedMul(radius, diry)

        local spark = P_SpawnMobj(x + offx, y + offy, z, MT_SPARKLETRAIL)

        local speed = N_RandomRange(6, 12)*FRACUNIT
        P_InstaThrust(spark, angle, speed)
        spark.momz = speed
        spark.colorized = true
        spark.color = color
    end
end

addHook("MobjThinker", function(mo)
    if not mo.valid then return end -- :skull:

    local onground = P_IsObjectOnGround(mo)

    if mo.prevonground == nil then
        mo.prevonground = onground
        return
    end

    -- I wanted to use MFE_JUSTHITFLOOR but it doesn't trigger sometimes :TailsSleep:
    if not mo.prevonground and onground and not mo.rainbow then
        local color = SKINCOLOR_WHITE

        if mo.target and mo.target.player then
            color = mo.target.player.skincolor
        end

        spawnPoof(mo.x, mo.y, mo.z, color)
    end

    if onground then
        if mo.rainbow then
            mo.rainbow = false

            spawnSparks(mo.x, mo.y, mo.z)
        end
    end

    mo.prevonground = onground
end, MT_BANANA)

addHook("PlayerSpin", function(player, inflictor, source)
    if inflictor and inflictor.type == MT_BANANA and inflictor.health > 1 then
        local color = SKINCOLOR_WHITE
        local colorizecolor = nil

        if inflictor.target and inflictor.target.player then
            color = inflictor.target.player.skincolor
            colorizecolor = color
        end

        spawnPoof(inflictor.x, inflictor.y, inflictor.z, color)

        if player.mo and colorizecolor then
            player.mo.bananasnipecolor = colorizecolor
        end
    end
end)

addHook("PlayerThink", function(p)
    if not (p.mo and p.mo.bananasnipecolor ~= nil) then return end

    if p.kartstuff[k_spinouttimer] or p.kartstuff[k_wipeoutslow] then
        p.mo.color = p.mo.bananasnipecolor
        p.mo.colorized = true
    else
        p.mo.bananasnipecolor = nil
        p.mo.color = p.skincolor
        p.mo.colorized = false
    end
end)
