-- Item duration indicator
-- Shows duration of grow, hyudoro, invincible, rocket sneakers, and sneakers on the bottom of the screen
-- Originally found in KL_BoostChaining_Speedboost_1.pk3 by Home
-- Forked and improved by Indev for Sunflower's Garden
-- Added HP mod support, moved the timers slightly, added a cvar to disable it, and more
local cv_showitemtimers = CONFIG_RegisterVar({
    name = "showitemtimers",
    defaultvalue = "Yes",
    possiblevalue = CV_YesNo,

	config_menu = "Item Timers",
	displayname = "Enable",
	description = "Show Item Timers.",
})

local cv_slipstreamcharge = CONFIG_RegisterVar({
    name = "showslipstreamcharge",
    defaultvalue = "Text",
    possiblevalue = { Off = 0, Text = 1, Arrows = 2, },

	config_menu = "Item Timers",
	displayname = "Show Slipstream Charge",
	description = "Show slipstream charge above timers.",
})

local cv_dn_hidetimers = CONFIG_RegisterVar({
    name = "hidetimersinnitro",
    defaultvalue = "Yes",
    possiblevalue = CV_YesNo,

	config_menu = "Item Timers",
	displayname = "Hide in Drift Nitro",
	description = "Hide item timers when Drift Nitro is active.",
})

local cv_timersfollowchar = CONFIG_RegisterVar({
    name = "timersfollowchar",
    defaultvalue = "No",
    possiblevalue = CV_YesNo,

	config_menu = "Item Timers",
	displayname = "Follow Character",
	description = "Draw timers next to character instead of fixed position.",
})

-- TODO - implement config_hudmove for these. Rn wouldn't really work because for timers to draw you need a buff
local cv_xoffset = CONFIG_RegisterVar({
    name = "timers_xoffset",
    defaultvalue = 0,

	config_menu = "Item Timers",
	displayname = "Item Timers X Offset",
	description = "Item Timers X Offset",
})

local cv_yoffset = CONFIG_RegisterVar({
    name = "timers_yoffset",
    defaultvalue = 0,

	config_menu = "Item Timers",
	displayname = "Item Timers Y Offset",
	description = "Item Timers Y Offset",
})

local cv_sort = CONFIG_RegisterVar({
    name = "timers_sort",
    defaultvalue = "Time",
    possiblevalue = { Time = 1, Name = 2 },

	config_menu = "Item Timers",
	displayname = "Sort By",
	description = "Change ordering of timers.",
})

local cv_sort_invert = CONFIG_RegisterVar({
    name = "timers_sort_invert",
    defaultvalue = "Off",
    possiblevalue = CV_OnOff,

	config_menu = "Item Timers",
	displayname = "Invert order",
	description = "Invert ordering of timers.",
})

local sortfuncs = {
    [1] = function(a, b, cmp) return cmp(a.timer, b.timer) end,
    [2] = function(a, b, cmp) return cmp(a.name, b.name) end,
}

local function gt(a, b) return a > b end
local function lt(a, b) return a < b end

local function sort(a, b)
    return sortfuncs[cv_sort.value](a, b, cv_sort_invert.value and lt or gt)
end

local havespb = false

local function setSpbTimer(p, value)
    if p.spbtimer == nil then
        p.spbtimer = value
    else
        p.spbtimer = min(p.spbtimer, value)
    end
end

local function getBestRank()
    local best = 255

    for player in players.iterate do
        if not (player.spectator or player.exiting) and player.kartstuff[k_position] < best then
            best = player.kartstuff[k_position]
        end
    end

    return best
end

local function SpbTimer(actor)
    havespb = true

    if actor.tracer and actor.tracer.player then
        local p = actor.tracer.player
        local bestrank = getBestRank()

        if p.kartstuff[k_position] > bestrank then
            setSpbTimer(p, actor.extravalue2)

            for player in players.iterate do
                if player.kartstuff[k_position] == bestrank then
                    setSpbTimer(player, actor.extravalue2)
                elseif player ~= p then
                    player.spbtimer = nil
                end
            end
        else
            for player in players.iterate do
                player.spbtimer = nil
            end
        end
    end
end

addHook("MobjThinker", SpbTimer, MT_SPB)

addHook("ThinkFrame", function()
    if havespb then
        havespb = false -- Spb thinker have to set this to true this every frame
        return
    end

    for player in players.iterate do
        player.spbtimer = nil
    end
end)

local cache = {}

local function cachePatches(v, name, patches)
    cache[name] = {}

    for i = 1, #patches do
        table.insert(cache[name], v.cachePatch(patches[i]))
    end

    return cache[name]
end

local cv_ss_chargetoboost
local function lookupCvars()
    if not cv_ss_chargetoboost then cv_ss_chargetoboost = CV_FindVar("slipstream_chargetoboost") end
end

addHook("MapLoad", lookupCvars)
addHook("NetVars", lookupCvars)

local function drawSlipstreamChargeText(v, x, y, p, flags, tiny)
    if not cv_ss_chargetoboost then return end

    local strtype = tiny and "small" or "thin"

    local fullstr = "SLIPSTREAM"

    local len = min(p.slipstreamcharge / (cv_ss_chargetoboost.value / #fullstr), #fullstr)

    if len == 0 then return end

    local str = fullstr:sub(1, len)

    if p.slipstreamboost then
        if not SG_Color2Chat then
            str = "\x83"..str
        else
            local skincolor = p.mo and p.mo.valid and p.mo.color or p.skincolor

            str = SG_Color2Chat[skincolor]..str
        end
    end

    v.drawString(x-v.stringWidth(str, flags, strtype)/2, y, str, flags, strtype)
end

local sschargegfx
local function drawSlipstreamChargeArrows(v, x, y, p, flags, tiny)
    if not cv_ss_chargetoboost then return end

	if sschargegfx == nil then
		sschargegfx = {}

		for i = 0, 9 do
			sschargegfx[i] = v.cachePatch(string.format("SLIPSTR%d", i))
		end
	end

    local len = min(p.slipstreamcharge / (cv_ss_chargetoboost.value / 9), 9)

	if len == 0 then return end

	len = len - 1

	local patch
	local color = SKINCOLOR_GREY
	local scale = tiny and FRACUNIT/2 or FRACUNIT

	if len < 7 then
		patch = sschargegfx[len]
	else
		patch = sschargegfx[7 + (leveltime/6) % 3]

		if p.slipstreamboost then
			if p.mo then
				color = p.mo.color
			else
				color = p.skincolor
			end
		end
	end

	v.drawScaled((x-patch.width/2)*FRACUNIT, y*FRACUNIT, scale, patch, flags, v.getColormap(TC_DEFAULT, color))
end

local drawsscharge = {
	[1] = drawSlipstreamChargeText,
	[2] = drawSlipstreamChargeArrows,
}
local function drawSlipstreamCharge(v, x, y, p, flags, tiny)
	drawsscharge[cv_slipstreamcharge.value](v, x, y, p, flags, tiny)
end

local function startBoost(p,b)
	if b == 0 then
		if leveltime < 270 or not ringsOn then
			return p.kartstuff[k_startboost]
		elseif leveltime >= 270 then
			return 0
		end
	elseif b == 1 then
		if leveltime < 270 or not ringsOn then
			return 0
		elseif leveltime >= 270 and ringsOn then
			return p.kartstuff[k_startboost]
		end

	end

end

hud.add(function(v, p, c)
    -- Driftnitro has its own slipstream indicator
	--if driftmodData and driftmodData.isOn then return end -- Not anymore :Trollic:
    if not p.spectator and cv_slipstreamcharge.value then
		local ofs = 0

		if driftmodData and driftmodData.isOn then
			ofs = (splitscreen > 0) and -4 or -30
		end

        local x = 158
        local y = 164 + ofs
        local flags = V_SNAPTOBOTTOM

        local splitnum = c.pnum - 1

        if splitscreen == 1 then
            if splitnum == 0 then
                y = 64 + ofs
                flags = 0
            end
        elseif splitscreen > 1 then
            if splitnum % 2 == 0 then -- p1 and p3
                x = 60
            else
                x = 260
            end

            if splitnum < 2 then -- p1 and p2
                y = 72
                flags = 0
            else
                y = 172
            end
        end

        if p.slipstreamcharge ~= nil and p.slipstreamcharge > 0 then
            drawSlipstreamCharge(v, x, y, p, flags, splitscreen > 1)
        end
    end
end)

local function getTimeString(timer)
    local seconds = G_TicsToSeconds(timer)
    local centiseconds = G_TicsToCentiseconds(timer)

    return seconds .. "." .. (centiseconds < 9 and "0" or "") .. centiseconds
end

hud.add(function(v, p, c)

	if not p.spectator and cv_showitemtimers.value and not (cv_dn_hidetimers.value and driftmodData and driftmodData.isOn) then

        local bestrank = getBestRank()

		-- name - for cache
        -- timer - timer to use
        -- patches - list of patch names
        -- anim_frames - frames for 1 animation step (used only for invincibility, in fact)
		local timerTable = {
			{
                name = "shoe",
                timer = (not (driftmodData and driftmodData.isOn) and p.kartstuff[k_sneakertimer]) or 0,
                patches = {"K_ISSHOE"},
                anim_frames = 1,
            },
			{
                name = "invincible",
                timer = p.kartstuff[k_invincibilitytimer],
                patches = {"K_ISINV1", "K_ISINV2", "K_ISINV3", "K_ISINV4", "K_ISINV5", "K_ISINV6"},
                anim_frames = 3,
            },
			{
                name = "grow",
                timer = p.kartstuff[k_growshrinktimer],
                patches = {"K_ISGROW"},
                anim_frames = 1,
            },
			{
                name = "rocketsneakers",
                timer = p.kartstuff[k_rocketsneakertimer],
                patches = {"K_ISRSHE"},
                anim_frames = 1
            },
			{
                name = "hyudoro",
                timer = p.kartstuff[k_hyudorotimer],
                patches = {"K_ISHYUD"},
                anim_frames = 1,
            },
			{
                name = "stackedboost",
                timer = p.stackedboosttimer,
                patches = {"CHNICON"},
                anim_frames = 1,
            },
			{
				name = "driftsparkboost",
				timer = p.kartstuff[k_driftboost],
				patches = {"DRSP1", "DRSP2"},
				anim_frames = 2,
			},
			{
				name = "startboost",
				timer = startBoost(p,0),
				patches = {"K_ISSTB"},
				anim_frames = 1,
			},
			{
				name = "ringboost",
				timer = startBoost(p,1),
				patches = {"K_ISRNG"},
				anim_frames = 1,
			},
            {
                name = "gnomeboost",
                timer = p.mo and p.mo.gnomeboost or 0,
                patches = {"GNHTA0"},
                anim_frames = 1,
                offsetx = 16,
                offsety = 18,
            }
		}
		-- sort table

		table.sort(timerTable, function(a, b) if(a.timer ~= nil and b.timer ~= nil) then return sort(a, b) end end)

        local has_sneaker = p.kartstuff[k_itemtype] == KITEM_SNEAKER or p.kartstuff[k_itemtype] == KITEM_ROCKETSNEAKER or p.kartstuff[k_rocketsneakertimer]
        table.insert(timerTable, {
            name = "firmsneaker",
            timer = has_sneaker and p.FSpositiontimer or 0,
            patches = {"K_ISFIRM"},
            anim_frames = 1,
        })

        -- This one always should be at end
        if p.mo and p.mo.valid and not (driftmodData and driftmodData.isOn) then
            local stack = 0
            if booststack and booststack.running then
                stack = booststack.stackedGradeSum(p.mo)
            elseif p.stackedboosts ~= nil then
                stack = p.stackedboosts
            end
            table.insert(timerTable, {
                name = "stackedboost",
                timer = stack,
                patches = {"DRSP1", "DRSP2"},
                anim_frames = 2,
            })
        end

        -- Ditto
        table.insert(timerTable, {
            name = "spbtimer",
            timer = p.spbtimer,
            patches = {"K_ISSPB"},
            anim_frames = 1,
        })

        -- ^
        table.insert(timerTable, {
            name = "spinout",
            timer = max(max(p.kartstuff[k_spinouttimer], p.kartstuff[k_wipeoutslow]), p.kartstuff[k_squishedtimer]),
            patches = {"DIZZA0", "DIZZB0", "DIZZC0", "DIZZD0"},
            anim_frames = 3,
        })

        local iconX = 150
		local iconY = 170
        local iconFlags = V_SNAPTOBOTTOM
        local splitnum = c.pnum - 1
        local stepX = 30
        local dup = cv_timersfollowchar.value and v.dupx() or 1

        if hpmod and hpmod.running then iconX = 240 end
        if driftmodData and driftmodData.isOn then iconX = 80 end
		if ringsOn then iconX = 190 end

        if splitscreen == 1 then
            if splitnum == 0 then
                iconY = 70
                iconFlags = 0
            end
        elseif splitscreen > 1 then
            stepX = 26 -- Make them a bit more compact

            if splitnum % 2 == 0 then -- p1 and p3
                iconX = 50
            else -- p2 and p4
                iconX = 250
            end

            if splitnum < 2 then -- p1 and p2
                iconY = 70
                iconFlags = 0
            else -- p3 and p4
                iconY = 170
            end

            -- Those offsets look fine in all those gamemodes, how convinient :3
            if hpmod and hpmod.running or
               driftmodData and driftmodData.isOn or
               ringsOn then

                if splitnum % 2 == 0 then
                    iconX = 100
                else
                    iconX = 200
                end
            end
        end

        iconFlags = $|V_HUDTRANS

        if cv_timersfollowchar.value then
            iconFlags = $|V_NOSCALESTART
            iconX, iconY = SG_GetScreenCoords(v, p, c, p.mo, p.mo.height + 40*mapobjectscale, true)

            if iconX == nil then return end

            iconX = $>>FRACBITS
            iconY = $>>FRACBITS
        end

        iconX = $ + cv_xoffset.value*dup
        iconY = $ + cv_yoffset.value*dup

		local iconXOffset = -stepX*dup
		for i, icon in ipairs(timerTable) do
			if(icon.timer and icon.timer > 0) then
				iconXOffset = $ + stepX*dup
			end
        end

		iconX = $ - (iconXOffset/2)
		local hasBeenOffset = false
		for i, icon in ipairs(timerTable) do

			-- Draw icon/relevant timer
			if icon.timer and icon.timer > 0 then
				local timer = icon.timer

				local timerstring = getTimeString(timer)

                local patches = cache[icon.name]

                if patches == nil then
                    patches = cachePatches(v, icon.name, icon.patches)
                end

                local patch_num = (leveltime % (icon.anim_frames * #patches) / icon.anim_frames) + 1

				local cmap = nil

				if icon.name == "driftsparkboost" then
					if icon.timer > 50 then
						cmap = v.getColormap(TC_DEFAULT, (1+leveltime%(MAXSKINCOLORS-1)))
					elseif icon.timer > 20 then
						cmap = v.getColormap(TC_DEFAULT, SKINCOLOR_KETCHUP)
					else
						cmap = v.getColormap(TC_DEFAULT, SKINCOLOR_SAPPHIRE)
					end
                elseif icon.name == "gnomeboost" then
                    cmap = v.getColormap(TC_DEFAULT, p.skincolor)
                end

                if icon.name == "stackedboost" then
                    timerstring = tostring(icon.timer)
                end

                if icon.name == "spinout" or icon.name == "spbtimer" and p.kartstuff[k_position] == getBestRank() then
                    timerstring = "\133"..timerstring
                end

				v.drawScaled((iconX + (icon.offsetx or 0)*dup)*FRACUNIT, (iconY + (icon.offsety or 0)*dup)*FRACUNIT, 6*FRACUNIT/10, patches[patch_num], iconFlags, cmap)

                if icon.name ~= "firmsneaker" or not SG_CanHaveFirmsneaker(p) then
				    v.drawString(iconX + 7*dup, iconY + 15*dup, timerstring, iconFlags, "thin")
                end

				iconX = $ + stepX*dup
			end


		end

	end
end)
